# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os

import fixtures
from oslo_config import cfg
from oslo_messaging import conffixture as messaging_conffixture
from oslo_utils import timeutils
from oslotest import base

from karbor.common import config  # noqa Need to register global_opts
from karbor.db import migration
from karbor.db.sqlalchemy import api as sqla_api
from karbor import rpc
from karbor.tests.unit import conf_fixture


CONF = cfg.CONF

_DB_CACHE = None


class Database(fixtures.Fixture):

    def __init__(self, db_api, db_migrate, sql_connection):
        super(Database, self).__init__()
        self.sql_connection = sql_connection

        # Suppress logging for test runs
        migrate_logger = logging.getLogger('migrate')
        migrate_logger.setLevel(logging.WARNING)

        self.engine = db_api.get_engine()
        self.engine.dispose()
        conn = self.engine.connect()
        db_migrate.db_sync()

        self._DB = "".join(line for line in conn.connection.iterdump())
        self.engine.dispose()

    def setUp(self):
        super(Database, self).setUp()

        conn = self.engine.connect()
        conn.connection.executescript(self._DB)
        self.addCleanup(self.engine.dispose)


class TestCase(base.BaseTestCase):

    """Test case base class for all unit tests."""

    def setUp(self):
        """Run before each test method to initialize test environment."""
        super(TestCase, self).setUp()

        rpc.add_extra_exmods("karbor.tests.unit")
        self.addCleanup(rpc.clear_extra_exmods)
        self.addCleanup(rpc.cleanup)

        self.messaging_conf = messaging_conffixture.ConfFixture(CONF)
        self.messaging_conf.transport_driver = 'fake'
        self.messaging_conf.response_timeout = 15
        self.useFixture(self.messaging_conf)
        rpc.init(CONF)

        conf_fixture.set_defaults(CONF)
        CONF([], default_config_files=[])

        # NOTE(vish): We need a better method for creating fixtures for tests
        #             now that we have some required db setup for the system
        #             to work properly.
        self.start = timeutils.utcnow()

        CONF.set_default('connection', 'sqlite://', 'database')
        CONF.set_default('sqlite_synchronous', False, 'database')

        global _DB_CACHE
        if not _DB_CACHE:
            _DB_CACHE = Database(sqla_api, migration,
                                 sql_connection=CONF.database.connection)
        self.useFixture(_DB_CACHE)

        self.override_config('policy_file',
                             os.path.join(
                                 os.path.abspath(
                                     os.path.join(
                                         os.path.dirname(__file__),
                                         '..',
                                     )
                                 ),
                                 'tests/unit/policy.json'),
                             group='oslo_policy')

    def override_config(self, name, override, group=None):
        """Cleanly override CONF variables."""
        CONF.set_override(name, override, group)
        self.addCleanup(CONF.clear_override, name, group)

    def flags(self, **kw):
        """Override CONF variables for a test."""
        for k, v in kw.items():
            self.override_config(k, v)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from karbor.resource import Resource
from karbor.services.protection import provider
from karbor.tests import base
from karbor.tests.unit.protection import fakes
from oslo_config import cfg

CONF = cfg.CONF

(
    parent_type,
    child_type,
    grandchild_type,
) = fakes.FakeProtectionPlugin.SUPPORTED_RESOURCES

parent = Resource(id='A1', name='parent', type=parent_type)
child = Resource(id='B1', name='child', type=child_type)
grandchild = Resource(id='C1', name='grandchild', type=grandchild_type)
resource_graph = {
    parent: [child],
    child: [grandchild],
    grandchild: [],
}


class ProviderRegistryTest(base.TestCase):
    def setUp(self):
        super(ProviderRegistryTest, self).setUp()

    @mock.patch.object(provider.PluggableProtectionProvider, '_load_bank')
    @mock.patch.object(provider.PluggableProtectionProvider,
                       '_register_plugin')
    def test_load_providers(self, mock_load_bank, mock_register_plugin):
        pr = provider.ProviderRegistry()
        self.assertEqual(1, mock_register_plugin.call_count)
        self.assertEqual(1, mock_load_bank.call_count)
        self.assertEqual(1, len(pr.providers))

        self.assertEqual('fake_provider1', pr.providers['fake_id1'].name)
        self.assertNotIn('fake_provider2', pr.providers)

    def test_provider_bank_config(self):
        pr = provider.ProviderRegistry()
        provider1 = pr.show_provider('fake_id1')
        self.assertEqual('thor', provider1.bank._plugin.fake_host)

    def test_provider_plugin_config(self):
        pr = provider.ProviderRegistry()
        provider1 = pr.show_provider('fake_id1')
        plugins = provider1.load_plugins()
        self.assertEqual('user', plugins['Test::ResourceA'].fake_user)

    def test_list_provider(self):
        pr = provider.ProviderRegistry()
        self.assertEqual(1, len(pr.list_providers()))

    def test_show_provider(self):
        pr = provider.ProviderRegistry()
        provider_list = pr.list_providers()
        for provider_node in provider_list:
            self.assertTrue(pr.show_provider(provider_node['id']))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
import greenlet

from datetime import datetime
from datetime import timedelta
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils

from karbor.common import constants
from karbor import context
from karbor import objects
from karbor.services.operationengine.engine.executors import base


green_thread_executor_opts = [
    cfg.IntOpt('max_concurrent_operations',
               default=0,
               help='number of maximum concurrent running operations,'
                    '0 means no hard limit'
               )
]

CONF = cfg.CONF
CONF.register_opts(green_thread_executor_opts, 'operationengine')

LOG = logging.getLogger(__name__)


class GreenThreadExecutor(base.BaseExecutor):

    def __init__(self, operation_manager):
        super(GreenThreadExecutor, self).__init__(operation_manager)
        self._operation_thread_map = {}

    def execute_operation(self, operation_id, triggered_time,
                          expect_start_time, window_time, **kwargs):

        if operation_id in self._operation_thread_map:
            LOG.warning("Execute operation(%s), the previous one has not been"
                        " finished", operation_id)
            return

        num = CONF.operationengine.max_concurrent_operations
        if num and len(self._operation_thread_map) >= num:
            LOG.warning("The amount of concurrent running operations "
                        "exceeds %d", num)
            return
        self._operation_thread_map[operation_id] = None

        end_time_for_run = expect_start_time + timedelta(seconds=window_time)
        ret = self._update_operation_state(
            operation_id,
            {'state': constants.OPERATION_STATE_RUNNING,
             'end_time_for_run': end_time_for_run})
        if not ret:
            self._operation_thread_map.pop(operation_id, None)
            return

        if operation_id not in self._operation_thread_map:
            # This function is invoked by trigger which may runs in the
            # green thread. So if operation_id is not exist, it may be
            # canceled by 'cancel_operation' during the call to DB in
            # the codes above.
            LOG.warning("Operation(%s) is not exist after call to DB",
                        operation_id)
            return

        param = {
            'operation_id': operation_id,
            'triggered_time': triggered_time,
            'expect_start_time': expect_start_time,
            'window_time': window_time,
            'run_type': constants.OPERATION_RUN_TYPE_EXECUTE
        }
        try:
            self._create_thread(self._run_operation, operation_id, param)
        except Exception:
            self._operation_thread_map.pop(operation_id, None)
            LOG.exception("Execute operation (%s), and create green thread "
                          "failed", operation_id)

    def cancel_operation(self, operation_id):
        gt = self._operation_thread_map.get(operation_id, None)
        if gt is not None:  # can not use 'if gt' instead
            # If the thead has not started, it will be killed;
            # else, it will run until finishes its work.
            gt.cancel()
        else:
            self._operation_thread_map.pop(operation_id, None)

    def resume_operation(self, operation_id, **kwargs):
        end_time = kwargs.get('end_time_for_run')
        now = datetime.utcnow()
        if not isinstance(end_time, datetime) or now > end_time:
            return

        window = int(timeutils.delta_seconds(now, end_time))
        param = {
            'operation_id': operation_id,
            'triggered_time': now,
            'expect_start_time': now,
            'window_time': window,
            'run_type': constants.OPERATION_RUN_TYPE_RESUME
        }
        self._create_thread(self._run_operation, operation_id, param)

    def shutdown(self):
        for op_id, gt in self._operation_thread_map.items():
            if gt is None:
                continue

            gt.cancel()
            try:
                gt.wait()  # wait untile the thread finishes its work
            except (greenlet.GreenletExit, Exception):
                pass

        self._operation_thread_map = {}

    def _run_operation(self, operation_id, param):

        try:
            try:
                operation = objects.ScheduledOperation.get_by_id(
                    context.get_admin_context(), operation_id)
            except Exception:
                LOG.exception("Run operation(%s), get operation failed",
                              operation_id)
                return

            try:
                param['user_id'] = operation.user_id
                param['project_id'] = operation.project_id

                self._operation_manager.run_operation(
                    operation.operation_type,
                    operation.operation_definition,
                    param=param)
            except Exception:
                LOG.exception("Run operation(%s) failed", operation_id)

        finally:
            self._update_operation_state(
                operation_id,
                {'state': constants.OPERATION_STATE_REGISTERED})

    def _update_operation_state(self, operation_id, updates):

        ctxt = context.get_admin_context()
        try:
            state_ref = objects.ScheduledOperationState.get_by_operation_id(
                ctxt, operation_id)
            for item, value in updates.items():
                setattr(state_ref, item, value)
            state_ref.save()
        except Exception:
            LOG.exception("Execute operation(%s), update state failed",
                          operation_id)
            return False
        return True

    def _on_gt_done(self, gt, *args, **kwargs):
        op_id = args[0]
        try:
            del self._operation_thread_map[op_id]
        except Exception:
            LOG.warning("Unknown operation id(%s) received, "
                        "when the green thread exit", op_id)

    def _create_thread(self, function, operation_id, param):
        gt = eventlet.spawn(function, operation_id, param)
        self._operation_thread_map[operation_id] = gt
        gt.link(self._on_gt_done, operation_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

try:
    from unittest import mock
except ImportError:
    import mock
from oslo_config import cfg
from oslo_db import exception as db_exc

from karbor.cmd import api as karbor_api
from karbor.cmd import manage as karbor_manage
from karbor.tests import base
from karbor import version

CONF = cfg.CONF


class TestKarborApiCmd(base.TestCase):
    """Unit test cases for python modules under karbor/cmd."""

    def setUp(self):
        super(TestKarborApiCmd, self).setUp()
        sys.argv = ['karbor-api']
        CONF(sys.argv[1:], project='karbor', version=version.version_string())

    def tearDown(self):
        super(TestKarborApiCmd, self).tearDown()

    @mock.patch('karbor.service.WSGIService')
    @mock.patch('karbor.service.process_launcher')
    @mock.patch('karbor.rpc.init')
    @mock.patch('oslo_log.log.setup')
    def test_main(self, log_setup, rpc_init, process_launcher,
                  wsgi_service):
        launcher = process_launcher.return_value
        server = wsgi_service.return_value
        server.workers = mock.sentinel.worker_count

        karbor_api.main()

        self.assertEqual('karbor', CONF.project)
        self.assertEqual(CONF.version, version.version_string())
        log_setup.assert_called_once_with(CONF, "karbor")
        rpc_init.assert_called_once_with(CONF)
        process_launcher.assert_called_once_with()
        wsgi_service.assert_called_once_with('osapi_karbor')
        launcher.launch_service.assert_called_once_with(server,
                                                        workers=server.workers)
        launcher.wait.assert_called_once_with()


class TestKarborManageCmd(base.TestCase):
    """Unit test cases for python modules under karbor/cmd/manage.py."""

    def setUp(self):
        super(TestKarborManageCmd, self).setUp()
        sys.argv = ['karbor-manage']
        CONF(sys.argv[1:], project='karbor', version=version.version_string())

    def tearDown(self):
        super(TestKarborManageCmd, self).tearDown()

    def test_db_commands_upgrade_out_of_range(self):
        version = 1111111111
        db_cmds = karbor_manage.DbCommands()
        exit = self.assertRaises(SystemExit, db_cmds.sync, version + 1)
        self.assertEqual(1, exit.code)

    @mock.patch("oslo_db.sqlalchemy.migration.db_sync")
    def test_db_commands_script_not_present(self, db_sync):
        db_sync.side_effect = db_exc.DbMigrationError
        db_cmds = karbor_manage.DbCommands()
        exit = self.assertRaises(SystemExit, db_cmds.sync, 101)
        self.assertEqual(1, exit.code)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from karbor.context import RequestContext
from karbor.services.protection.clients import swift
from karbor.tests import base


class SwiftClientTest(base.TestCase):
    def setUp(self):
        super(SwiftClientTest, self).setUp()
        service_catalog = [
            {
                'endpoints': [
                    {'publicURL': 'http://127.0.0.1:8080/v1/AUTH_abcd', }
                ],
                'type': 'object-store',
                'name': 'swift',
            },
        ]
        self._context = RequestContext(user_id='demo',
                                       project_id='abcd',
                                       auth_token='efgh',
                                       service_catalog=service_catalog)

        self.conf = cfg.ConfigOpts()
        swift.register_opts(self.conf)

    def test_create_client_by_keystone(self):
        self.skipTest('test needs revision')
        auth_url = 'http://127.0.0.1/identity/'
        self.conf.set_default('swift_auth_url',
                              auth_url,
                              'swift_client')
        self.conf.set_override('swift_user', 'demo', 'swift_client')
        self.conf.set_override('swift_key', 'secrete', 'swift_client')
        self.conf.set_override('swift_tenant_name', 'abcd', 'swift_client')
        sc = swift.create(self._context, self.conf)
        self.assertEqual(sc.authurl, auth_url)
        self.assertEqual('demo', sc.user)
        self.assertEqual('secrete', sc.key)
        self.assertEqual('abcd', sc.os_options['tenant_name'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from karbor.common import constants
from karbor.services.protection import resource_flow
from oslo_log import log as logging
from taskflow import task

LOG = logging.getLogger(__name__)


class InitiateDeleteTask(task.Task):
    def execute(self, checkpoint, *args, **kwargs):
        LOG.debug("Initiate delete checkpoint_id: %s", checkpoint.id)
        checkpoint.status = constants.CHECKPOINT_STATUS_DELETING
        checkpoint.commit()

    def revert(self, checkpoint, *args, **kwargs):
        LOG.debug("Failed to delete checkpoint_id: %s", checkpoint.id)
        checkpoint.status = constants.CHECKPOINT_STATUS_ERROR_DELETING
        checkpoint.commit()


class CompleteDeleteTask(task.Task):
    def execute(self, checkpoint):
        LOG.debug("Complete delete checkpoint_id: %s", checkpoint.id)
        checkpoint.delete()


def get_flow(context, workflow_engine, checkpoint, provider):
    LOG.info("Start get checkpoint flow, checkpoint_id: %s", checkpoint.id)
    flow_name = "Delete_Checkpoint_" + checkpoint.id
    delete_flow = workflow_engine.build_flow(flow_name, 'linear')
    resource_graph = checkpoint.resource_graph
    plugins = provider.load_plugins()
    resources_task_flow = resource_flow.build_resource_flow(
        operation_type=constants.OPERATION_DELETE,
        context=context,
        workflow_engine=workflow_engine,
        resource_graph=resource_graph,
        plugins=plugins,
        parameters=None
    )
    workflow_engine.add_tasks(
        delete_flow,
        InitiateDeleteTask(),
        resources_task_flow,
        CompleteDeleteTask(),
    )
    flow_engine = workflow_engine.get_engine(delete_flow,
                                             store={'checkpoint': checkpoint})
    return flow_engine



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_utils import timeutils

from karbor import objects
from karbor.tests.unit import objects as test_objects

NOW = timeutils.utcnow().replace(microsecond=0)

CheckpointRecord_ID = '36ea41b2-c358-48a7-9117-70cb7617410a'

Fake_CheckpointRecord = {
    'created_at': NOW,
    'deleted_at': None,
    'updated_at': NOW,
    'deleted': False,
    "id": CheckpointRecord_ID,
    "project_id": "586cc6ce-e286-40bd-b2b5-dd32694d9944",
    "checkpoint_id": "2220f8b1-975d-4621-a872-fa9afb43cb6c",
    "checkpoint_status": "available",
    "provider_id": "39bb894794b741e982bd26144d2949f6",
    "plan_id": "efc6a88b-9096-4bb6-8634-cda182a6e12b",
    "operation_id": "64e51e85-4f31-441f-9a5d-6e93e3196628",
    "create_by": "operation-engine",
    "extend_info": "[{"
                    "'id': '0354ca9d-dcd0-46b6-9334-0d78759fd275',"
                    "'type': 'OS::Nova::Server',"
                    "'name': 'vm1'"
                    "}]"
}


class TestCheckpointRecord(test_objects.BaseObjectsTestCase):
    CheckpointRecord_Class = objects.CheckpointRecord

    @mock.patch('karbor.db.checkpoint_record_get')
    def test_get_by_id(self, checkpoint_record_get):
        db_checkpoint_record = Fake_CheckpointRecord.copy()
        checkpoint_record_get.return_value = db_checkpoint_record

        checkpoint_record = self.CheckpointRecord_Class.get_by_id(
            self.context,
            CheckpointRecord_ID)
        self._compare(self, db_checkpoint_record, checkpoint_record)
        checkpoint_record_get.assert_called_once_with(
            self.context,
            CheckpointRecord_ID)

    @mock.patch('karbor.db.checkpoint_record_create')
    def test_create(self, checkpoint_record_create):
        db_checkpoint_record = Fake_CheckpointRecord.copy()
        checkpoint_record_create.return_value = db_checkpoint_record

        checkpoint_record = self.CheckpointRecord_Class(context=self.context)
        checkpoint_record.create()
        self._compare(self, db_checkpoint_record, checkpoint_record)
        checkpoint_record_create.assert_called_once_with(self.context, {})

    @mock.patch('karbor.db.checkpoint_record_update')
    def test_save(self, checkpoint_record_update):
        db_checkpoint_record = Fake_CheckpointRecord
        checkpoint_record = self.CheckpointRecord_Class._from_db_object(
            self.context,
            self.CheckpointRecord_Class(),
            db_checkpoint_record)
        checkpoint_record.checkpoint_status = 'error'
        checkpoint_record.save()

        checkpoint_record_update.assert_called_once_with(
            self.context,
            checkpoint_record.id,
            {'checkpoint_status': 'error'})

    @mock.patch('karbor.db.checkpoint_record_destroy')
    def test_destroy(self, checkpoint_record_destroy):
        db_checkpoint_record = Fake_CheckpointRecord
        checkpoint_record = self.CheckpointRecord_Class._from_db_object(
            self.context,
            self.CheckpointRecord_Class(),
            db_checkpoint_record)
        checkpoint_record.destroy()
        checkpoint_record_destroy.assert_called_once_with(
            self.context,
            checkpoint_record.id)



# Copyright 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import namedtuple

Stack = namedtuple('Stack', ['id',
                             'stack_status',
                             'stack_name'])

FakeStacks = {}


class FakeHeatClient(object):
    class Stacks(object):
        def create(self, stack_name, template):
            stack = Stack(id='fake_stack_id',
                          stack_name=stack_name,
                          stack_status='CREATE_IN_PROGRESS')
            FakeStacks[stack.id] = stack
            return {
                'stack': {
                    'id': 'stack_id_1',
                }
            }

        def get(self, stack_id):
            return FakeStacks[stack_id]

    def __init__(self):
        super(FakeHeatClient, self).__init__()
        self.stacks = self.Stacks()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from glanceclient import client as gc
from oslo_config import cfg
from oslo_log import log as logging

from karbor.common import config
from karbor.services.protection.clients import utils

LOG = logging.getLogger(__name__)

SERVICE = 'glance'
glance_client_opts = [
    cfg.StrOpt(SERVICE + '_endpoint',
               help='URL of the glance endpoint.'),
    cfg.StrOpt(SERVICE + '_catalog_info',
               default='image:glance:publicURL',
               help='Info to match when looking for glance in the service '
               'catalog. Format is: separated values of the form: '
               '<service_type>:<service_name>:<endpoint_type> - '
               'Only used if glance_endpoint is unset'),
    cfg.StrOpt(SERVICE + '_ca_cert_file',
               help='Location of the CA certificate file '
                    'to use for client requests in SSL connections.'),
    cfg.BoolOpt(SERVICE + '_auth_insecure',
                default=False,
                help='Bypass verification of server certificate when '
                     'making SSL connection to Glance.'),
]

CONFIG_GROUP = '%s_client' % SERVICE
CONF = cfg.CONF
CONF.register_opts(config.service_client_opts, group=CONFIG_GROUP)
CONF.register_opts(glance_client_opts, group=CONFIG_GROUP)
CONF.set_default('service_name', 'glance', CONFIG_GROUP)
CONF.set_default('service_type', 'image', CONFIG_GROUP)

GLANCECLIENT_VERSION = '2'


def create(context, conf, **kwargs):
    conf.register_opts(glance_client_opts, group=CONFIG_GROUP)

    client_config = conf[CONFIG_GROUP]
    url = utils.get_url(SERVICE, context, client_config, **kwargs)
    LOG.debug("Creating glance client with url %s.", url)

    if kwargs.get('session'):
        return gc.Client(GLANCECLIENT_VERSION, session=kwargs.get('session'),
                         endpoint=url)

    args = {
        'endpoint': url,
        'token': context.auth_token,
        'cacert': client_config.glance_ca_cert_file,
        'insecure': client_config.glance_auth_insecure,
    }
    return gc.Client(GLANCECLIENT_VERSION, **args)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import namedtuple
import mock

from karbor.common import constants
from karbor.context import RequestContext
from karbor import resource
from karbor.services.protection.protectable_plugins.network import \
    NetworkProtectablePlugin
from karbor.tests import base
from keystoneauth1 import session as keystone_session
from neutronclient.v2_0 import client
from oslo_config import cfg

CONF = cfg.CONF

server_info = namedtuple('server_info',
                         field_names=['id', 'type', 'name', 'addresses'])
project_info = namedtuple('project_info', field_names=['id', 'type'])

FakePorts = {'ports': [
    {'fixed_ips': [{'subnet_id': 'subnet-1',
                    'ip_address': '10.0.0.21'}],
     'id': 'port-1',
     'mac_address': 'mac_address_1',
     'device_id': 'vm_id_1',
     'name': '',
     'admin_state_up': True,
     'network_id': '658e1063-4ee3-4649-a2c9'},
    {'fixed_ips': [{'subnet_id': 'subnet-1',
                    'ip_address': '10.0.0.22'}],
     'id': 'port-2',
     'mac_address': 'mac_address_2',
     'device_id': 'vm_id_2',
     'name': '',
     'admin_state_up': True,
     'network_id': 'network_id_2'}
]}


class NetworkProtectablePluginTest(base.TestCase):
    def setUp(self):
        super(NetworkProtectablePluginTest, self).setUp()

        service_catalog = [{
            'type': 'network',
            'endpoints': [{'publicURL': 'http://127.0.0.1:9696'}]
        }, {
            'type': 'compute',
            'endpoints': [{'publicURL': 'http://127.0.0.1:8774/v2.1/abcd'}]
        }]
        self._context = RequestContext(user_id='admin',
                                       project_id='abcd',
                                       auth_token='efgh',
                                       service_catalog=service_catalog)

    @mock.patch('karbor.services.protection.client_factory.ClientFactory.'
                '_generate_session')
    def test_create_client_by_endpoint(self, mock_generate_session):
        CONF.set_default('neutron_endpoint', 'http://127.0.0.1:9696',
                         'neutron_client')
        CONF.set_default('nova_endpoint', 'http://127.0.0.1:8774/v2.1',
                         'nova_client')
        plugin = NetworkProtectablePlugin(self._context)
        neutronclient = plugin._neutron_client(self._context)
        novaclient = plugin._nova_client(self._context)
        mock_generate_session.return_value = keystone_session.Session(
            auth=None)
        self.assertEqual('network',
                         neutronclient.httpclient.service_type)
        self.assertEqual('http://127.0.0.1:9696',
                         neutronclient.httpclient.endpoint_url)
        self.assertEqual('http://127.0.0.1:8774/v2.1/abcd',
                         novaclient.client.management_url)

    @mock.patch('karbor.services.protection.client_factory.ClientFactory.'
                '_generate_session')
    def test_create_client_by_catalog(self, mock_generate_session):
        CONF.set_default('neutron_catalog_info', 'network:neutron:publicURL',
                         'neutron_client')
        CONF.set_default('nova_catalog_info', 'compute:nova:publicURL',
                         'nova_client')
        plugin = NetworkProtectablePlugin(self._context)
        neutronclient = plugin._neutron_client(self._context)
        novaclient = plugin._nova_client(self._context)
        mock_generate_session.return_value = keystone_session.Session(
            auth=None)
        self.assertEqual('network',
                         neutronclient.httpclient.service_type)
        self.assertEqual('http://127.0.0.1:9696',
                         neutronclient.httpclient.endpoint_url)
        self.assertEqual('http://127.0.0.1:8774/v2.1/abcd',
                         novaclient.client.management_url)

    def test_get_resource_type(self):
        plugin = NetworkProtectablePlugin(self._context)
        self.assertEqual(constants.NETWORK_RESOURCE_TYPE,
                         plugin.get_resource_type())

    def test_get_parent_resource_type(self):
        plugin = NetworkProtectablePlugin(self._context)
        self.assertItemsEqual(plugin.get_parent_resource_types(),
                              (constants.SERVER_RESOURCE_TYPE,
                               constants.PROJECT_RESOURCE_TYPE))

    @mock.patch.object(client.Client, 'list_networks')
    def test_list_resources(self, mock_client_list_networks):
        plugin = NetworkProtectablePlugin(self._context)

        fake_network_info = {'networks': [
            {u'status': u'ACTIVE',
             u'description': u'',
             u'tenant_id': u'abcd',
             u'name': u'private'},
            {u'status': u'ACTIVE',
             u'description': u'',
             u'name': u'ext_net',
             u'tenant_id': u'abcd'}
            ]}

        mock_client_list_networks.return_value = fake_network_info
        self.assertEqual(plugin.list_resources(self._context),
                         [resource.Resource
                             (type=constants.NETWORK_RESOURCE_TYPE,
                              id='abcd',
                              name="Network Topology")])

    @mock.patch.object(client.Client, 'list_networks')
    def test_get_project_dependent_resources(self, mock_client_list_networks):
        project = project_info(id='abcd',
                               type=constants.PROJECT_RESOURCE_TYPE)
        plugin = NetworkProtectablePlugin(self._context)
        fake_network_info = {'networks': [
            {u'status': u'ACTIVE',
             u'description': u'',
             u'tenant_id': u'abcd',
             u'name': u'private'},
            {u'status': u'ACTIVE',
             u'description': u'',
             u'name': u'ext_net',
             u'tenant_id': u'abcd'}
            ]}
        mock_client_list_networks.return_value = fake_network_info
        self.assertEqual(plugin.get_dependent_resources(self._context,
                                                        project),
                         [resource.Resource
                             (type=constants.NETWORK_RESOURCE_TYPE,
                              id='abcd',
                              name="Network Topology")])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The restores api."""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils

from webob import exc

import karbor
from karbor.api import common
from karbor.api.openstack import wsgi
from karbor.common import constants
from karbor import exception
from karbor.i18n import _

from karbor import objects
from karbor.objects import base as objects_base
import karbor.policy
from karbor.services.protection import api as protection_api
from karbor import utils

import six

query_restore_filters_opt = cfg.ListOpt(
    'query_restore_filters',
    default=['status'],
    help="Restore filter options which "
    "non-admin user could use to "
    "query restores. Default values "
    "are: ['status']")
CONF = cfg.CONF
CONF.register_opt(query_restore_filters_opt)

LOG = logging.getLogger(__name__)


def check_policy(context, action, target_obj=None):
    target = {
        'project_id': context.project_id,
        'user_id': context.user_id,
    }

    if isinstance(target_obj, objects_base.KarborObject):
        # Turn object into dict so target.update can work
        target.update(
            target_obj.obj_to_primitive() or {})
    else:
        target.update(target_obj or {})

    _action = 'restore:%s' % action
    karbor.policy.enforce(context, _action, target)


class RestoreViewBuilder(common.ViewBuilder):
    """Model a server API response as a python dictionary."""

    _collection_name = "restores"

    def detail(self, request, restore):
        """Detailed view of a single restore."""
        restore_ref = {
            'restore': {
                'id': restore.get('id'),
                'project_id': restore.get('project_id'),
                'provider_id': restore.get('provider_id'),
                'checkpoint_id': restore.get('checkpoint_id'),
                'restore_target': restore.get('restore_target'),
                'parameters': restore.get('parameters'),
                'status': restore.get('status'),
                'resources_status': restore.get('resources_status'),
                'resources_reason': restore.get('resources_reason'),
            }
        }
        return restore_ref

    def detail_list(self, request, restores, restore_count=None):
        """Detailed view of a list of restores."""
        return self._list_view(self.detail, request, restores,
                               restore_count,
                               self._collection_name)

    def _list_view(self, func, request, restores, restore_count,
                   coll_name=_collection_name):
        """Provide a view for a list of restores.

        :param func: Function used to format the restore data
        :param request: API request
        :param restores: List of restores in dictionary format
        :param restore_count: Length of the original list of restores
        :param coll_name: Name of collection, used to generate the next link
                          for a pagination query
        :returns: restore data in dictionary format
        """
        restores_list = [func(request, restore)['restore']
                         for restore in restores]
        restores_links = self._get_collection_links(request,
                                                    restores,
                                                    coll_name,
                                                    restore_count)
        restores_dict = {
            'restores': restores_list
        }
        if restores_links:
            restores_dict['restores_links'] = restores_links

        return restores_dict


class RestoresController(wsgi.Controller):
    """The Restores API controller for the OpenStack API."""

    _view_builder_class = RestoreViewBuilder

    def __init__(self):
        self.protection_api = protection_api.API()
        super(RestoresController, self).__init__()

    def show(self, req, id):
        """Return data about the given restore."""
        context = req.environ['karbor.context']

        LOG.info("Show restore with id: %s", id, context=context)

        if not uuidutils.is_uuid_like(id):
            msg = _("Invalid restore id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        try:
            restore = self._restore_get(context, id)
        except exception.RestoreNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)

        LOG.info("Show restore request issued successfully.",
                 resource={'id': restore.id})
        return self._view_builder.detail(req, restore)

    def index(self, req):
        """Returns a list of restores, transformed through view builder."""
        context = req.environ['karbor.context']

        LOG.info("Show restore list", context=context)

        params = req.params.copy()
        marker, limit, offset = common.get_pagination_params(params)
        sort_keys, sort_dirs = common.get_sort_params(params)
        filters = params

        utils.remove_invalid_filter_options(
            context,
            filters,
            self._get_restore_filter_options())

        utils.check_filters(filters)
        restores = self._get_all(context, marker, limit,
                                 sort_keys=sort_keys,
                                 sort_dirs=sort_dirs,
                                 filters=filters,
                                 offset=offset)

        retval_restores = self._view_builder.detail_list(req, restores)

        LOG.info("Show restore list request issued successfully.")

        return retval_restores

    def _get_all(self, context, marker=None, limit=None, sort_keys=None,
                 sort_dirs=None, filters=None, offset=None):
        check_policy(context, 'get_all')

        if filters is None:
            filters = {}

        all_tenants = utils.get_bool_param('all_tenants', filters)

        try:
            if limit is not None:
                limit = int(limit)
                if limit <= 0:
                    msg = _('limit param must be positive')
                    raise exception.InvalidInput(reason=msg)
        except ValueError:
            msg = _('limit param must be an integer')
            raise exception.InvalidInput(reason=msg)

        if filters:
            LOG.debug("Searching by: %s.", six.text_type(filters))

        if context.is_admin and all_tenants:
            # Need to remove all_tenants to pass the filtering below.
            del filters['all_tenants']
            restores = objects.RestoreList.get_all(
                context, marker, limit,
                sort_keys=sort_keys,
                sort_dirs=sort_dirs,
                filters=filters,
                offset=offset)
        else:
            restores = objects.RestoreList.get_all_by_project(
                context, context.project_id, marker, limit,
                sort_keys=sort_keys, sort_dirs=sort_dirs, filters=filters,
                offset=offset)

        LOG.info("Get all restores completed successfully.")
        return restores

    def _get_restore_filter_options(self):
        """Return restores search options allowed by non-admin."""
        return CONF.query_restore_filters

    def create(self, req, body):
        """Creates a new restore."""
        if not self.is_valid_body(body, 'restore'):
            raise exc.HTTPUnprocessableEntity()

        LOG.debug('Create restore request body: %s', body)
        context = req.environ['karbor.context']
        check_policy(context, 'create')
        restore = body['restore']
        LOG.debug('Create restore request : %s', restore)

        if not restore.get("provider_id"):
            msg = _("provider_id must be provided when creating "
                    "a restore.")
            raise exception.InvalidInput(reason=msg)

        if not restore.get("checkpoint_id"):
            msg = _("checkpoint_id must be provided when creating "
                    "a restore.")
            raise exception.InvalidInput(reason=msg)

        parameters = restore.get("parameters")
        if not isinstance(parameters, dict):
            msg = _("parameters must be a dict when creating"
                    " a restore.")
            raise exception.InvalidInput(reason=msg)

        # restore_auth and restore_target are optional
        # Heat client can be initialized using current login tenant when the
        # restore_target and restore_auth is not provided.
        restore_auth = restore.get("restore_auth")
        if restore_auth is not None:
            if not isinstance(restore_auth, dict):
                msg = _("restore_auth must be a dict when creating"
                        " a restore.")
                raise exception.InvalidInput(reason=msg)

        restore_properties = {
            'project_id': context.project_id,
            'provider_id': restore.get('provider_id'),
            'checkpoint_id': restore.get('checkpoint_id'),
            'restore_target': restore.get('restore_target'),
            'parameters': parameters,
            'status': constants.RESTORE_STATUS_IN_PROGRESS,
        }

        restoreobj = objects.Restore(context=context,
                                     **restore_properties)
        restoreobj.create()
        LOG.debug('call restore RPC  : restoreobj:%s', restoreobj)

        # call restore rpc API of protection service
        try:
            self.protection_api.restore(context, restoreobj, restore_auth)
        except Exception:
            # update the status of restore
            update_dict = {
                "status": constants.RESTORE_STATUS_FAILURE
            }
            check_policy(context, 'update', restoreobj)
            restoreobj = self._restore_update(context,
                                              restoreobj.get("id"),
                                              update_dict)

        retval = self._view_builder.detail(req, restoreobj)

        return retval

    def _restore_get(self, context, restore_id):
        if not uuidutils.is_uuid_like(restore_id):
            msg = _("Invalid restore id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        restore = objects.Restore.get_by_id(context, restore_id)
        try:
            check_policy(context, 'get', restore)
        except exception.PolicyNotAuthorized:
            # raise RestoreNotFound instead to make sure karbor behaves
            # as it used to
            raise exception.RestoreNotFound(restore_id=restore_id)
        LOG.info("Restore info retrieved successfully.")
        return restore

    def _restore_update(self, context, restore_id, fields):
        try:
            restore = self._restore_get(context, restore_id)
        except exception.RestoreNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)

        if isinstance(restore, objects_base.KarborObject):
            restore.update(fields)
            restore.save()
            LOG.info("restore updated successfully.")
            return restore
        else:
            msg = _("The parameter restore must be a object of "
                    "KarborObject class.")
            raise exception.InvalidInput(reason=msg)


def create_resource():
    return wsgi.Resource(RestoresController())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
from karbor.common import constants
from karbor.context import RequestContext
from karbor.resource import Resource
from karbor.services.protection.bank_plugin import Bank
from karbor.services.protection.bank_plugin import BankPlugin
from karbor.services.protection.bank_plugin import BankSection
from karbor.services.protection import client_factory
from karbor.services.protection.protection_plugins.network \
    import network_plugin_schemas
from karbor.services.protection.protection_plugins.network. \
    neutron_protection_plugin import NeutronProtectionPlugin
from karbor.tests import base
import mock
from oslo_config import cfg

FakeNetworks = {'networks': [
    {u'status': u'ACTIVE',
     u'router:external': False,
     u'availability_zone_hints': [],
     u'availability_zones': [u'nova'],
     u'ipv4_address_scope': None,
     u'description': u'',
     u'provider:physical_network': None,
     u'subnets': [u'129f1bc5-4282-4f9d-ae60-4db1e1cac22d',
                  u'0b42e051-5a33-4ac4-9a4f-691d0891d760'],
     u'updated_at': u'2016-04-23T05:07:06',
     u'tenant_id': u'f6f6d0b2591f41acb8257656d70029fc',
     u'created_at': u'2016-04-23T05:07:06',
     u'tags': [],
     u'ipv6_address_scope': None,
     u'provider:segmentation_id': 1057,
     u'provider:network_type': u'vxlan',
     u'port_security_enabled': True,
     u'admin_state_up': True,
     u'shared': False,
     u'mtu': 1450,
     u'id': u'9b68fb64-39d4-4d41-8cc9-f27846c6e5f5',
     u'name': u'private'},

    {u'provider:physical_network': None,
     u'ipv6_address_scope': None,
     u'port_security_enabled': True,
     u'provider:network_type': u'local',
     u'id': u'49ef013d-9bb2-4b8f-9eea-e45563efc420',
     u'router:external': True,
     u'availability_zone_hints': [],
     u'availability_zones': [u'nova'],
     u'ipv4_address_scope': None,
     u'shared': False,
     u'status': u'ACTIVE',
     u'subnets': [u'808c3b3f-3d79-4c5b-a5b6-95dd07abeb2d'],
     u'description': u'',
     u'tags': [],
     u'updated_at': u'2016-04-25T07:14:53',
     u'is_default': False,
     u'provider:segmentation_id': None,
     u'name': u'ext_net',
     u'admin_state_up': True,
     u'tenant_id': u'f6f6d0b2591f41acb8257656d70029fc',
     u'created_at': u'2016-04-25T07:14:53',
     u'mtu': 1500}
    ]}

FakeSubnets = {'subnets': [
    {u'description': u'',
     u'enable_dhcp': True,
     u'network_id': u'49ef013d-9bb2-4b8f-9eea-e45563efc420',
     u'tenant_id': u'f6f6d0b2591f41acb8257656d70029fc',
     u'created_at': u'2016-04-25T07:15:25',
     u'dns_nameservers': [],
     u'updated_at': u'2016-04-25T07:15:25',
     u'ipv6_ra_mode': None,
     u'allocation_pools': [{u'start': u'192.168.21.2',
                            u'end': u'192.168.21.254'}],
     u'gateway_ip': u'192.168.21.1',
     u'ipv6_address_mode': None,
     u'ip_version': 4,
     u'host_routes': [],
     u'cidr': u'192.168.21.0/24',
     u'id': u'808c3b3f-3d79-4c5b-a5b6-95dd07abeb2d',
     u'subnetpool_id': None,
     u'name': u'ext_subnet'},
    ]}

FakePorts = {'ports': [
    {u'allowed_address_pairs': [],
     u'extra_dhcp_opts': [],
     u'updated_at': u'2016-04-25T07:15:59',
     u'device_owner':
     u'network:router_gateway',
     u'port_security_enabled': False,
     u'binding:profile': {},
     u'fixed_ips': [{u'subnet_id': u'808c3b3f-3d79-4c5b-a5b6-95dd07abeb2d',
                     u'ip_address': u'192.168.21.3'}],
     u'id': u'2b34c97a-4ccc-44c0-bc50-b7bbfc3508eb',
     u'security_groups': [],
     u'binding:vif_details': {},
     u'binding:vif_type': u'unbound',
     u'mac_address': u'fa:16:3e:00:47:f2',
     u'status': u'DOWN',
     u'binding:host_id': u'',
     u'description': u'',
     u'device_id': u'7fc86d4b-4c0e-4ed8-8d39-e27b7c1b7ae8',
     u'name': u'',
     u'admin_state_up': True,
     u'network_id': u'49ef013d-9bb2-4b8f-9eea-e45563efc420',
     u'dns_name': None,
     u'created_at': u'2016-04-25T07:15:59',
     u'binding:vnic_type': u'normal',
     u'tenant_id': u''},
    ]}

FakeRoutes = {'routers': [
    {u'status': u'ACTIVE',
     u'external_gateway_info':
        {u'network_id': u'49ef013d-9bb2-4b8f-9eea-e45563efc420',
         u'enable_snat': True,
         u'external_fixed_ips':
            [{u'subnet_id': u'808c3b3f-3d79-4c5b-a5b6-95dd07abeb2d',
             u'ip_address': u'192.168.21.3'}
             ]},
     u'availability_zone_hints': [],
     u'availability_zones': [],
     u'description': u'',
     u'admin_state_up': True,
     u'tenant_id': u'f6f6d0b2591f41acb8257656d70029fc',
     u'distributed': False,
     u'routes': [],
     u'ha': False,
     u'id': u'7fc86d4b-4c0e-4ed8-8d39-e27b7c1b7ae8',
     u'name': u'provider_route'}
    ]}

FakeSecGroup = {'security_groups': [
    {u'tenant_id': u'23b119d06168447c8dbb4483d9567bd8',
     u'name': u'default',
     u'id': u'97910ed4-1dcb-4704-8814-3ddca818ac16',
     u'description': u'Default security group',
     u'security_group_rules': [
            {u'remote_group_id': u'ac4a6134-0176-44db-abab-559d284c4cdc',
              u'direction': u'ingress',
              u'protocol': None,
              u'description': u'',
              u'ethertype': u'IPv4',
              u'remote_ip_prefix': None,
              u'port_range_max': None,
              u'security_group_id': u'ac4a6134-0176-44db-abab-559d284c4cdc',
              u'port_range_min': None,
              u'tenant_id': u'23b119d06168447c8dbb4483d9567bd8',
              u'id': u'21416a24-6a7a-4830-bbec-1426b21e085a'},

            {u'remote_group_id': u'ac4a6134-0176-44db-abab-559d284c4cdc',
             u'direction': u'ingress',
             u'protocol': None,
             u'description': u'',
             u'ethertype': u'IPv6',
             u'remote_ip_prefix': None,
             u'port_range_max': None,
             u'security_group_id': u'ac4a6134-0176-44db-abab-559d284c4cdc',
             u'port_range_min': None,
             u'tenant_id': u'23b119d06168447c8dbb4483d9567bd8',
             u'id': u'47f67d6a-4e73-465a-9f4d-d9b850f85f22'},

            {u'remote_group_id': None,
             u'direction': u'egress',
             u'protocol': None,
             u'description': u'',
             u'ethertype': u'IPv6',
             u'remote_ip_prefix': None,
             u'port_range_max': None,
             u'security_group_id': u'ac4a6134-0176-44db-abab-559d284c4cdc',
             u'port_range_min': None,
             u'tenant_id': u'23b119d06168447c8dbb4483d9567bd8',
             u'id': u'c24e7148-820c-4147-9032-6fcdb96db6f7'}]},
    ]}


def call_hooks(operation, checkpoint, resource, context, parameters, **kwargs):
    def noop(*args, **kwargs):
        pass

    hooks = (
        'on_prepare_begin',
        'on_prepare_finish',
        'on_main',
        'on_complete',
    )
    for hook_name in hooks:
        hook = getattr(operation, hook_name, noop)
        hook(checkpoint, resource, context, parameters, **kwargs)


class FakeNeutronClient(object):
    def list_networks(self):
        return FakeNetworks

    def list_subnets(self):
        return FakeSubnets

    def list_ports(self):
        return FakePorts

    def list_routers(self):
        return FakeRoutes

    def list_security_groups(self):
        return FakeSecGroup


class FakeBankPlugin(BankPlugin):
    def __init__(self):
        self._objects = {}

    def create_object(self, key, value):
        self._objects[key] = value

    def update_object(self, key, value):
        self._objects[key] = value

    def get_object(self, key):
        value = self._objects.get(key, None)
        if value is None:
            raise Exception
        return value

    def list_objects(self, prefix=None, limit=None, marker=None):
        objects_name = []
        if prefix is not None:
            for key, value in self._objects.items():
                if key.find(prefix) == 0:
                    objects_name.append(key.lstrip(prefix))
        else:
            objects_name = self._objects.keys()
        return objects_name

    def delete_object(self, key):
        self._objects.pop(key)

    def get_owner_id(self):
        return

fake_checkpointid = "checkpoint_id"
fake_project_id = "abcd"
fake_bank = Bank(FakeBankPlugin())
fake_bank_section = BankSection(bank=fake_bank, section="fake")

ResourceNode = collections.namedtuple(
    "ResourceNode",
    ["value"]
)


class CheckpointCollection(object):
    def __init__(self):
        self.bank_section = fake_bank_section

    def get_resource_bank_section(self, resource_id):
        return self.bank_section


class NeutronProtectionPluginTest(base.TestCase):
    def setUp(self):
        super(NeutronProtectionPluginTest, self).setUp()

        self.plugin = NeutronProtectionPlugin()

        cfg.CONF.set_default('neutron_endpoint',
                             'http://127.0.0.1:9696',
                             'neutron_client')

        self.cntxt = RequestContext(user_id='admin',
                                    project_id='abcd',
                                    auth_token='efgh')

        self.neutron_client = client_factory.ClientFactory.create_client(
            "neutron", self.cntxt)
        self.checkpoint = CheckpointCollection()

    def test_get_options_schema(self):
        options_schema = self.plugin.get_options_schema(
            'OS::Neutron::Network')
        self.assertEqual(options_schema,
                         network_plugin_schemas.OPTIONS_SCHEMA)

    def test_get_restore_schema(self):
        options_schema = self.plugin.get_restore_schema(
            'OS::Neutron::Network')
        self.assertEqual(options_schema,
                         network_plugin_schemas.RESTORE_SCHEMA)

    def test_get_saved_info_schema(self):
        options_schema = self.plugin.get_saved_info_schema(
            'OS::Neutron::Network')
        self.assertEqual(options_schema,
                         network_plugin_schemas.SAVED_INFO_SCHEMA)

    def test_get_supported_resources_types(self):
        types = self.plugin.get_supported_resources_types()
        self.assertEqual([constants.NETWORK_RESOURCE_TYPE], types)

    @mock.patch('karbor.services.protection.clients.neutron.create')
    def test_create_backup(self, mock_neutron_create):
        resource = Resource(id="network_id_1",
                            type=constants.NETWORK_RESOURCE_TYPE,
                            name="test")

        fake_bank_section.update_object = mock.MagicMock()

        protect_operation = self.plugin.get_protect_operation(resource)
        mock_neutron_create.return_value = self.neutron_client

        self.neutron_client.list_networks = mock.MagicMock()
        self.neutron_client.list_networks.return_value = FakeNetworks

        self.neutron_client.list_subnets = mock.MagicMock()
        self.neutron_client.list_subnets.return_value = FakeSubnets

        self.neutron_client.list_ports = mock.MagicMock()
        self.neutron_client.list_ports.return_value = FakePorts

        self.neutron_client.list_routers = mock.MagicMock()
        self.neutron_client.list_routers.return_value = FakeRoutes

        self.neutron_client.list_security_groups = mock.MagicMock()
        self.neutron_client.list_security_groups.return_value = FakeSecGroup

        call_hooks(protect_operation, self.checkpoint, resource, self.cntxt,
                   {})

    @mock.patch('karbor.services.protection.clients.neutron.create')
    def test_delete_backup(self, mock_neutron_create):
        resource = Resource(id="network_id_1",
                            type=constants.NETWORK_RESOURCE_TYPE,
                            name="test")

        fake_bank._plugin._objects[
            "/resource_data/checkpoint_id/network_id_1/metadata"] = {
            "resource_id": "network_id_1",
            "network_metadata": {
                "id": "9b68fb64-39d4-4d41-8cc9-f27846c6e5f5",
                "router:external": False,
                "admin_state_up": True,
                "mtu": 1500
            },
            "subnet_metadata": {
                "id": "808c3b3f-3d79-4c5b-a5b6-95dd07abeb2d",
                "network_id": "9b68fb64-39d4-4d41-8cc9-f27846c6e5f5",
                "host_routes": [],
                "dns_nameservers": []
            },
            "port_metadata": {
                "id": "2b34c97a-4ccc-44c0-bc50-b7bbfc3508eb",
                "admin_state_up": True,
                "allowed_address_pairs": [],
                "fixed_ips": [{"subnet_id": "3d79-4c5b-a5b6-95dd07abeb2d",
                               "ip_address": "192.168.21.3"}]
            },
            "router_metadata": {
                "id": "4c0e-4ed8-8d39-e27b7c1b7ae8",
                "admin_state_up": True,
                "availability_zone_hints": [],
                "fixed_ips":  {"network_id": "9bb2-4b8f-9eea-e45563efc420",
                               "enable_snat": True
                               }
            },
            "security-group_metadata": {
                "id": "4ccc-44c0-bc50-b7bbfc3508eb",
                "description": "Default security group",
                "security_group_rules": []
            }
        }

        delete_operation = self.plugin.get_delete_operation(resource)
        mock_neutron_create.return_value = self.neutron_client

        call_hooks(delete_operation, self.checkpoint, resource, self.cntxt,
                   {})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Command-line flag library.

Emulates gflags by wrapping cfg.ConfigOpts.

The idea is to move fully to cfg eventually, and this wrapper is a
stepping stone.

"""

import socket

from oslo_config import cfg
from oslo_log import log as logging


CONF = cfg.CONF
logging.register_options(CONF)

core_opts = [
    cfg.StrOpt('state_path',
               default='/var/lib/karbor',
               deprecated_name='pybasedir',
               help="Top-level directory for maintaining karbor's state"),
]

debug_opts = [
]

CONF.register_cli_opts(core_opts)
CONF.register_cli_opts(debug_opts)

global_opts = [
    cfg.IntOpt('service_down_time',
               default=60,
               help='Maximum time since last check-in for a service to be '
                    'considered up'),
    cfg.StrOpt('operationengine_topic',
               default='karbor-operationengine',
               help='The topic that OperationEngine nodes listen on'),
    cfg.StrOpt('operationengine_manager',
               default='karbor.services.operationengine.manager.'
               'OperationEngineManager',
               help='Full class name for the Manager for OperationEngine'),
    cfg.StrOpt('protection_topic',
               default='karbor-protection',
               help='The topic that protection nodes listen on'),
    cfg.StrOpt('protection_manager',
               default='karbor.services.protection.manager.ProtectionManager',
               help='Full class name for the Manager for Protection'),
    cfg.HostAddressOpt('host',
                       default=socket.gethostname(),
                       help='Name of this node.  This can be an opaque '
                            'identifier. It is not necessarily a host '
                            'name, FQDN, or IP address.'),
    cfg.StrOpt('auth_strategy',
               default='keystone',
               choices=['noauth', 'keystone'],
               help='The strategy to use for auth. Supports noauth or '
                    'keystone.'),
]

CONF.register_opts(global_opts)


service_client_opts = [
    cfg.StrOpt('service_name',
               help='The name of service registered in Keystone'),

    cfg.StrOpt('service_type',
               help='The type of service registered in Keystone'),

    cfg.StrOpt('version',
               help='The version of service client'),

    cfg.StrOpt('region_id',
               default='RegionOne',
               help='The region id which the service belongs to.'),

    cfg.StrOpt('interface',
               default='internal',
               help='The network interface of the endpoint. Valid '
                    'values are: public, admin, internal.'),

    cfg.StrOpt('ca_cert_file',
               help='Location of the CA certificate file '
                    'to use for client requests in SSL connections.'),

    cfg.BoolOpt('auth_insecure',
                default=False,
                help='Bypass verification of server certificate when '
                     'making SSL connection to service.')
]


keystone_client_opts = [
    cfg.StrOpt('auth_uri',
               default='',
               help='Unversioned keystone url in format like '
                    'http://0.0.0.0:5000.')]


def list_opts():
    yield 'clients_keystone', keystone_client_opts


for group, opts in list_opts():
    CONF.register_opts(opts, group=group)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import service_token
from keystoneauth1 import session as keystone_session

import os
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

from karbor.common import karbor_keystone_plugin
from karbor import exception
from karbor.i18n import _

LOG = logging.getLogger(__name__)


class ClientFactory(object):
    _factory = None
    _keystone_plugin = None

    @staticmethod
    def _list_clients():
        clients_dir = os.path.join(os.path.dirname(__file__), 'clients')
        if not os.path.isdir(clients_dir):
            LOG.error('clients directory "%s" not found', clients_dir)
            return

        for file in os.listdir(clients_dir):
            name, ext = os.path.splitext(file)
            if name != '__init__' and name != 'utils' and ext == '.py':
                LOG.debug('Found client "%s"', name)
                yield '%s.clients.%s' % (__package__, name)

    @classmethod
    def _generate_session(cls, context, service, privileged_user=False):
        LOG.debug("Generate an auth session. privileged_user: %s",
                  privileged_user)
        plugin = cls.get_keystone_plugin()
        try:
            if privileged_user is True:
                auth_plugin = service_token.ServiceTokenAuthWrapper(
                    plugin.service_auth_plugin,
                    plugin.service_auth_plugin)
            else:
                auth_plugin = service_token.ServiceTokenAuthWrapper(
                    plugin.create_user_auth_plugin(context),
                    plugin.service_auth_plugin)
        except Exception:
            return None

        try:
            client_conf = cfg.CONF['%s_client' % service]
            auth_insecure = client_conf['%s_auth_insecure' % service]
            ca_file = client_conf['%s_ca_cert_file' % service]
            verify = False if auth_insecure else (ca_file or True)

        except Exception:
            verify = True

        return keystone_session.Session(auth=auth_plugin, verify=verify)

    @classmethod
    def get_keystone_plugin(cls):
        if not cls._keystone_plugin:
            cls._keystone_plugin = \
                karbor_keystone_plugin.KarborKeystonePlugin()
        return cls._keystone_plugin

    @classmethod
    def get_client_module(cls, service):
        if not cls._factory:
            cls._factory = {}
            for module in cls._list_clients():
                module = importutils.import_module(module)
                cls._factory[module.SERVICE] = module
        return cls._factory.get(service)

    @classmethod
    def create_client(cls, service, context, conf=cfg.CONF,
                      privileged_user=False, **kwargs):
        module = cls.get_client_module(service)
        if module is None:
            raise exception.KarborException(_('Unknown service(%s)') % service)

        kwargs['privileged_user'] = privileged_user
        kwargs['keystone_plugin'] = cls.get_keystone_plugin()
        if context or privileged_user:
            kwargs['session'] = cls._generate_session(context, service,
                                                      privileged_user)
        return module.create(context, conf, **kwargs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
from karbor.common import constants
from karbor.context import RequestContext
from karbor.resource import Resource
from karbor.services.protection.bank_plugin import Bank
from karbor.services.protection.bank_plugin import BankPlugin
from karbor.services.protection.bank_plugin import BankSection
from karbor.services.protection import client_factory
from karbor.services.protection.protection_plugins. \
    image.image_protection_plugin import GlanceProtectionPlugin
from karbor.services.protection.protection_plugins.image \
    import image_plugin_schemas
from karbor.tests import base
import mock
from oslo_config import cfg
from oslo_config import fixture


class FakeBankPlugin(BankPlugin):
    def update_object(self, key, value):
        return

    def get_object(self, key):
        return

    def list_objects(self, prefix=None, limit=None, marker=None,
                     sort_dir=None):
        return

    def delete_object(self, key):
        return

    def get_owner_id(self):
        return


fake_bank = Bank(FakeBankPlugin())
fake_bank_section = BankSection(bank=fake_bank, section="fake")

ResourceNode = collections.namedtuple(
    "ResourceNode",
    ["value",
     "child_nodes"]
)


Image = collections.namedtuple(
    "Image",
    ["disk_format",
     "container_format",
     "status"]
)


def call_hooks(operation, checkpoint, resource, context, parameters, **kwargs):
    def noop(*args, **kwargs):
        pass

    hooks = (
        'on_prepare_begin',
        'on_prepare_finish',
        'on_main',
        'on_complete',
    )
    for hook_name in hooks:
        hook = getattr(operation, hook_name, noop)
        hook(checkpoint, resource, context, parameters, **kwargs)


class CheckpointCollection(object):
    def __init__(self):
        super(CheckpointCollection, self).__init__()
        self.bank_section = fake_bank_section

    def get_resource_bank_section(self, resource_id):
        return self.bank_section


class GlanceProtectionPluginTest(base.TestCase):
    def setUp(self):
        super(GlanceProtectionPluginTest, self).setUp()

        plugin_config = cfg.ConfigOpts()
        plugin_config_fixture = self.useFixture(fixture.Config(plugin_config))
        plugin_config_fixture.load_raw_values(
            group='image_backup_plugin',
            poll_interval=0,
        )
        plugin_config_fixture.load_raw_values(
            group='image_backup_plugin',
            backup_image_object_size=65536,
        )
        self.plugin = GlanceProtectionPlugin(plugin_config)
        cfg.CONF.set_default('glance_endpoint',
                             'http://127.0.0.1:9292',
                             'glance_client')

        self.cntxt = RequestContext(user_id='demo',
                                    project_id='abcd',
                                    auth_token='efgh')
        self.glance_client = client_factory.ClientFactory.create_client(
            "glance", self.cntxt)
        self.checkpoint = CheckpointCollection()

    def test_get_options_schema(self):
        options_schema = self.plugin.get_options_schema(
            constants.IMAGE_RESOURCE_TYPE)
        self.assertEqual(options_schema, image_plugin_schemas.OPTIONS_SCHEMA)

    def test_get_restore_schema(self):
        options_schema = self.plugin.get_restore_schema(
            constants.IMAGE_RESOURCE_TYPE)
        self.assertEqual(options_schema, image_plugin_schemas.RESTORE_SCHEMA)

    def test_get_saved_info_schema(self):
        options_schema = self.plugin.get_saved_info_schema(
            constants.IMAGE_RESOURCE_TYPE)
        self.assertEqual(options_schema,
                         image_plugin_schemas.SAVED_INFO_SCHEMA)

    @mock.patch('karbor.services.protection.protection_plugins.image.'
                'image_protection_plugin.utils.status_poll')
    @mock.patch('karbor.services.protection.clients.glance.create')
    def test_create_backup(self, mock_glance_create, mock_status_poll):
        resource = Resource(id="123",
                            type=constants.IMAGE_RESOURCE_TYPE,
                            name='fake')

        fake_bank_section.update_object = mock.MagicMock()

        protect_operation = self.plugin.get_protect_operation(resource)
        mock_glance_create.return_value = self.glance_client

        self.glance_client.images.get = mock.MagicMock()
        self.glance_client.images.return_value = Image(
            disk_format="",
            container_format="",
            status="active"
        )

        fake_bank_section.update_object = mock.MagicMock()
        self.glance_client.images.data = mock.MagicMock()
        self.glance_client.images.data.return_value = []
        mock_status_poll.return_value = True
        call_hooks(protect_operation, self.checkpoint, resource, self.cntxt,
                   {})

    def test_delete_backup(self):
        resource = Resource(id="123",
                            type=constants.IMAGE_RESOURCE_TYPE,
                            name='fake')

        fake_bank_section.list_objects = mock.MagicMock()
        fake_bank_section.list_objects.return_value = ["data_1", "data_2"]
        fake_bank_section.delete_object = mock.MagicMock()
        delete_operation = self.plugin.get_delete_operation(resource)
        call_hooks(delete_operation, self.checkpoint, resource, self.cntxt,
                   {})

    def test_get_supported_resources_types(self):
        types = self.plugin.get_supported_resources_types()
        self.assertEqual([constants.IMAGE_RESOURCE_TYPE], types)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import OrderedDict
from copy import deepcopy
from oslo_utils import uuidutils

from karbor import exception
from karbor.services.protection.bank_plugin import Bank
from karbor.services.protection.bank_plugin import BankPlugin
from karbor.services.protection.bank_plugin import BankSection
from karbor.services.protection.bank_plugin import LeasePlugin
from karbor.tests import base


class _InMemoryBankPlugin(BankPlugin):
    def __init__(self, config=None):
        super(_InMemoryBankPlugin, self).__init__(config)
        self._data = OrderedDict()

    def update_object(self, key, value):
        self._data[key] = value

    def get_object(self, key):
        try:
            return deepcopy(self._data[key])
        except KeyError:
            raise exception.BankGetObjectFailed('no such object')

    def list_objects(self, prefix=None, limit=None, marker=None,
                     sort_dir=None):
        marker_found = marker is None
        for key in self._data.keys():
            if marker is not True and key != marker:
                if marker_found:
                    if prefix is None or key.startswith(prefix):
                        if limit is not None:
                            limit -= 1
                            if limit < 0:
                                return
                        yield key
            else:
                marker_found = True

    def delete_object(self, key):
        del self._data[key]

    def get_owner_id(self):
        return uuidutils.generate_uuid()


class _InMemoryLeasePlugin(LeasePlugin):

    def acquire_lease(self):
        pass

    def renew_lease(self):
        pass

    def check_lease_validity(self):
        return True


class BankSectionTest(base.TestCase):
    INVALID_PATHS = (
        '/',
        '/a$',
        '/path/',
        '/path/path/',
        'space space',
        '/path/../dots/',
        '/,',
    )

    VALID_PATHS = (
        '/key',
        '/top/key',
        '/top/middle/bottom/key',
        '/all_kinds/of-char.acters/@path1',
    )

    def _create_test_bank(self):
        return Bank(_InMemoryBankPlugin())

    def test_empty_key(self):
        bank = self._create_test_bank()
        section = BankSection(bank, "/prefix", is_writable=True)
        self.assertRaises(
            exception.InvalidParameterValue,
            section.update_object,
            "",
            "value",
        )
        self.assertRaises(
            exception.InvalidParameterValue,
            section.update_object,
            None,
            "value",
        )

    def test_update_invalid_object(self):
        bank = self._create_test_bank()
        for path in self.INVALID_PATHS:
            self.assertRaises(
                exception.InvalidParameterValue,
                bank.update_object,
                path,
                "value",
            )

    def test_get_invalid_object(self):
        bank = self._create_test_bank()
        for path in self.INVALID_PATHS:
            self.assertRaises(
                exception.InvalidParameterValue,
                bank.get_object,
                path,
            )

    def test_valid_object(self):
        value1 = 'value1'
        value2 = 'value2'
        bank = self._create_test_bank()
        for path in self.VALID_PATHS:
            bank.update_object(path, value1)
            bank.update_object(path, value2)
            res = bank.get_object(path)
            self.assertEqual(value2, res)
            bank.delete_object(path)
            self.assertRaises(
                exception.BankGetObjectFailed,
                bank.get_object,
                path,
            )

    def test_delete_object(self):
        bank = self._create_test_bank()
        section = BankSection(bank, "/prefix", is_writable=True)
        bank.update_object("/prefix/a", "value")
        bank.update_object("/prefix/b", "value")
        bank.update_object("/prefix/c", "value")
        section.delete_object("a")
        section.delete_object("/b")
        section.delete_object("//c")

    def test_list_objects(self):
        bank = self._create_test_bank()
        section = BankSection(bank, "/prefix", is_writable=True)
        bank.update_object("/prefix/KeyA", "value")
        bank.update_object("/prefix", "value")
        bank.update_object("/prefixKeyD", "value")  # Should not appear
        section.update_object("/KeyB", "value")
        section.update_object("KeyC", "value")
        expected_result = ["KeyA", "KeyB", "KeyC"]
        self.assertEqual(expected_result, list(section.list_objects("/")))
        self.assertEqual(expected_result, list(section.list_objects("///")))
        self.assertEqual(expected_result, list(section.list_objects(None)))
        self.assertEqual(expected_result, list(section.list_objects("Key")))
        self.assertEqual(
            expected_result[:2],
            list(section.list_objects("/", limit=2)))
        self.assertEqual(
            expected_result[2:4],
            list(section.list_objects("/", limit=2, marker="KeyB")))

    def test_read_only(self):
        bank = self._create_test_bank()
        section = BankSection(bank, "/prefix", is_writable=False)
        self.assertRaises(
            exception.BankReadonlyViolation,
            section.update_object,
            "object",
            "value",
        )
        bank.update_object("/prefix/object", "value")
        self.assertRaises(
            exception.BankReadonlyViolation,
            section.update_object,
            "object",
            "value",
        )
        self.assertRaises(
            exception.BankReadonlyViolation,
            section.delete_object,
            "object",
        )

    def test_double_dot_key(self):
        bank = self._create_test_bank()
        section = BankSection(bank, "/prefix")
        self.assertRaises(
            exception.InvalidParameterValue,
            section.update_object,
            "/../../",
            "",
        )

    def test_double_dot_section_prefix(self):
        bank = self._create_test_bank()
        self.assertRaises(
            exception.InvalidParameterValue,
            BankSection,
            bank,
            '/../../',
        )

    def test_nested_sections_get(self):
        bank = self._create_test_bank()
        top_section = BankSection(bank, "/top")
        mid_section = top_section.get_sub_section("/mid")
        bottom_section = mid_section.get_sub_section("/bottom")
        bottom_section.update_object("key", "value")
        self.assertEqual("value", bank.get_object("/top/mid/bottom/key"))
        self.assertEqual("value", bottom_section.get_object("key"))

    def test_nested_sections_list(self):
        bank = self._create_test_bank()
        top_section = BankSection(bank, "/top")
        mid_section = top_section.get_sub_section("/mid")
        bottom_section = mid_section.get_sub_section("/bottom")
        keys = ["KeyA", "KeyB", "KeyC"]
        for key in keys:
            bottom_section.update_object(key, "value")

        list_result = set(bottom_section.list_objects(prefix="Key"))
        self.assertEqual(set(keys), list_result)
        self.assertEqual(
            keys[:2],
            list(bottom_section.list_objects("/", limit=2)))
        self.assertEqual(
            keys[2:4],
            list(bottom_section.list_objects("/", limit=2, marker="KeyB")))

    def test_nested_sections_read_only(self):
        bank = self._create_test_bank()
        section = BankSection(bank, "/top", is_writable=False)
        self.assertRaises(
            exception.BankReadonlyViolation,
            section.get_sub_section,
            "/mid",
            is_writable=True,
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import mock

from iso8601 import iso8601
from oslo_utils import uuidutils
from oslo_versionedobjects import fields

from karbor import objects
from karbor.tests.unit import objects as test_objects


@objects.base.KarborObjectRegistry.register_if(False)
class TestObject(objects.base.KarborObject):
    fields = {
        'scheduled_at': objects.base.fields.DateTimeField(nullable=True),
        'uuid': objects.base.fields.UUIDField(),
        'text': objects.base.fields.StringField(nullable=True),
    }


class TestKarborObject(test_objects.BaseObjectsTestCase):
    """Tests methods from KarborObject."""

    def setUp(self):
        super(TestKarborObject, self).setUp()
        self.obj = TestObject(
            scheduled_at=None,
            uuid=uuidutils.generate_uuid(),
            text='text')
        self.obj.obj_reset_changes()

    def test_karbor_obj_get_changes_no_changes(self):
        self.assertEqual({}, self.obj.karbor_obj_get_changes())

    def test_karbor_obj_get_changes_other_changes(self):
        self.obj.text = 'text2'
        self.assertEqual({'text': 'text2'},
                         self.obj.karbor_obj_get_changes())

    def test_karbor_obj_get_changes_datetime_no_tz(self):
        now = datetime.datetime.utcnow()
        self.obj.scheduled_at = now
        self.assertEqual({'scheduled_at': now},
                         self.obj.karbor_obj_get_changes())

    def test_karbor_obj_get_changes_datetime_tz_utc(self):
        now_tz = iso8601.parse_date('2015-06-26T22:00:01Z')
        now = now_tz.replace(tzinfo=None)
        self.obj.scheduled_at = now_tz
        self.assertEqual({'scheduled_at': now},
                         self.obj.karbor_obj_get_changes())

    def test_karbor_obj_get_changes_datetime_tz_non_utc_positive(self):
        now_tz = iso8601.parse_date('2015-06-26T22:00:01+01')
        now = now_tz.replace(tzinfo=None) - datetime.timedelta(hours=1)
        self.obj.scheduled_at = now_tz
        self.assertEqual({'scheduled_at': now},
                         self.obj.karbor_obj_get_changes())

    def test_karbor_obj_get_changes_datetime_tz_non_utc_negative(self):
        now_tz = iso8601.parse_date('2015-06-26T10:00:01-05')
        now = now_tz.replace(tzinfo=None) + datetime.timedelta(hours=5)
        self.obj.scheduled_at = now_tz
        self.assertEqual({'scheduled_at': now},
                         self.obj.karbor_obj_get_changes())

    def test_refresh(self):
        @objects.base.KarborObjectRegistry.register_if(False)
        class MyTestObject(objects.base.KarborObject,
                           objects.base.KarborObjectDictCompat,
                           objects.base.KarborComparableObject):
            fields = {'id': fields.UUIDField(),
                      'name': fields.StringField()}

        test_obj = MyTestObject(id='1', name='foo')
        refresh_obj = MyTestObject(id='1', name='bar')
        with mock.patch(
                'karbor.objects.base.KarborObject.get_by_id') as get_by_id:
            get_by_id.return_value = refresh_obj

            test_obj.refresh()
            self._compare(self, refresh_obj, test_obj)

    def test_refresh_no_id_field(self):
        @objects.base.KarborObjectRegistry.register_if(False)
        class MyTestObjectNoId(objects.base.KarborObject,
                               objects.base.KarborObjectDictCompat,
                               objects.base.KarborComparableObject):
            fields = {'uuid': fields.UUIDField()}

        test_obj = MyTestObjectNoId(uuid='1', name='foo')
        self.assertRaises(NotImplementedError, test_obj.refresh)


class TestKarborComparableObject(test_objects.BaseObjectsTestCase):
    def test_comparable_objects(self):
        @objects.base.KarborObjectRegistry.register
        class MyComparableObj(objects.base.KarborObject,
                              objects.base.KarborObjectDictCompat,
                              objects.base.KarborComparableObject):
            fields = {'foo': fields.Field(fields.Integer())}

        class NonVersionedObject(object):
            pass

        obj1 = MyComparableObj(foo=1)
        obj2 = MyComparableObj(foo=1)
        obj3 = MyComparableObj(foo=2)
        obj4 = NonVersionedObject()
        self.assertTrue(obj1 == obj2)
        self.assertFalse(obj1 == obj3)
        self.assertFalse(obj1 == obj4)
        self.assertNotEqual(obj1, None)


class TestKarborDictObject(test_objects.BaseObjectsTestCase):
    @objects.base.KarborObjectRegistry.register_if(False)
    class TestDictObject(objects.base.KarborObjectDictCompat,
                         objects.base.KarborObject):
        obj_extra_fields = ['foo']

        fields = {
            'abc': fields.StringField(nullable=True),
            'def': fields.IntegerField(nullable=True),
        }

        @property
        def foo(self):
            return 42

    def test_dict_objects(self):
        obj = self.TestDictObject()
        self.assertIsNone(obj.get('non_existing'))
        self.assertEqual('val', obj.get('abc', 'val'))
        self.assertIsNone(obj.get('abc'))
        obj.abc = 'val2'
        self.assertEqual('val2', obj.get('abc', 'val'))
        self.assertEqual(42, obj.get('foo'))

        self.assertIn('foo', obj)
        self.assertIn('abc', obj)
        self.assertNotIn('def', obj)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Manage all operations.
"""

from karbor import exception
from karbor.i18n import _
from karbor.services.operationengine import operations


class OperationManager(object):
    """Manage all operation classes which are defined at operations dir."""
    def __init__(self, user_trust_manager):
        super(OperationManager, self).__init__()
        self._user_trust_manager = user_trust_manager
        all_ops = operations.all_operations()
        self._ops_map = {op.OPERATION_TYPE: op(self._user_trust_manager)
                         for op in all_ops}

    def _get_operation(self, operation_type):
        if operation_type not in self._ops_map:
            msg = (_("Invalid operation type: %s") % operation_type)
            raise exception.InvalidInput(msg)

        return self._ops_map[operation_type]

    def check_operation_definition(self, operation_type, operation_definition):
        """Check operation definition.

        :param operation_type: the type of operation
        :param operation_definition: the definition of operation
        :raise InvalidInput if the operation_type is invalid or
               InvalidOperationDefinition if operation_definition is invalid
        """
        op = self._get_operation(operation_type)
        op.check_operation_definition(operation_definition)

    def run_operation(self, operation_type, operation_definition, **kwargs):
        """Run operation.

        :param operation_type: the type of operation
        :param operation_definition: the definition of operation
        :raise InvalidInput if the operation_type is invalid.
        """
        op = self._get_operation(operation_type)
        op.run(operation_definition, **kwargs)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import botocore
import botocore.session
import logging
from oslo_config import cfg

LOG = logging.getLogger(__name__)
SERVICE = 's3'
s3_client_opts = [
    cfg.StrOpt(SERVICE + '_endpoint',
               help='URL of the S3 compatible storage endpoint.'),
    cfg.StrOpt(SERVICE + '_access_key',
               help='Access key for S3 compatible storage.'),
    cfg.StrOpt(SERVICE + '_secret_key',
               secret=True,
               help='Secret key for S3 compatible storage.'),
    cfg.IntOpt(SERVICE + '_retry_attempts',
               default=3,
               help='The number of retries to make for '
                    'S3 operations'),
    cfg.IntOpt(SERVICE + '_retry_backoff',
               default=2,
               help='The backoff time in seconds '
                    'between S3 retries')
]


def register_opts(conf):
    conf.register_opts(s3_client_opts, group=SERVICE + '_client')


def create(context, conf, **kwargs):
    register_opts(conf)

    client_config = conf.s3_client
    LOG.info('Creating s3 client with url %s.',
             client_config.s3_endpoint)
    return botocore.session.get_session().create_client(
        's3',
        aws_access_key_id=client_config.s3_access_key,
        aws_secret_access_key=client_config.s3_ecret_key,
        endpoint_url=client_config.s3_endpoint
    )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock
from oslo_config import cfg
from webob import exc

from karbor.api.v1 import restores
from karbor.common import constants
from karbor import context
from karbor import exception
from karbor.tests import base
from karbor.tests.unit.api import fakes

CONF = cfg.CONF

DEFAULT_PROJECT_ID = '39bb894794b741e982bd26144d2949f6'
DEFAULT_PROVIDER_ID = 'efc6a88b-9096-4bb6-8634-cda182a6e12a'
DEFAULT_CHECKPOINT_ID = '09edcbdc-d1c2-49c1-a212-122627b20968'
DEFAULT_RESTORE_TARGET = '192.168.1.2/identity/'
DEFAULT_RESTORE_AUTH = {
    'type': 'password',
    'username': 'demo',
    'password': 'test',
}
DEFAULT_PARAMETERS = {
}


class RestoreApiTest(base.TestCase):
    def setUp(self):
        super(RestoreApiTest, self).setUp()
        self.controller = restores.RestoresController()
        self.ctxt = context.RequestContext('demo', 'fakeproject', True)

    @mock.patch(
        'karbor.services.protection.api.API.restore')
    @mock.patch(
        'karbor.objects.restore.Restore.create')
    def test_restore_create(self, mock_restore_create,
                            mock_rpc_restore):
        restore = self._restore_in_request_body()
        body = {"restore": restore}
        req = fakes.HTTPRequest.blank('/v1/restores')
        self.controller.create(req, body)
        self.assertTrue(mock_restore_create.called)
        self.assertTrue(mock_rpc_restore.called)

    def test_restore_create_InvalidBody(self):
        restore = self._restore_in_request_body()
        body = {"restorexx": restore}
        req = fakes.HTTPRequest.blank('/v1/restores')
        self.assertRaises(exc.HTTPUnprocessableEntity, self.controller.create,
                          req, body)

    def test_restore_create_InvalidProviderId(self):
        restore = self._restore_in_request_body(provider_id="")
        body = {"restore": restore}
        req = fakes.HTTPRequest.blank('/v1/restores')
        self.assertRaises(exception.InvalidInput, self.controller.create,
                          req, body)

    def test_restore_create_Invalidcheckpoint_id(self):
        restore = self._restore_in_request_body(checkpoint_id="")
        body = {"restore": restore}
        req = fakes.HTTPRequest.blank('/v1/restores')
        self.assertRaises(exception.InvalidInput, self.controller.create,
                          req, body)

    @mock.patch(
        'karbor.api.v1.restores.RestoresController._get_all')
    def test_restore_list_detail(self, moak_get_all):
        req = fakes.HTTPRequest.blank('/v1/restores')
        self.controller.index(req)
        self.assertTrue(moak_get_all.called)

    @mock.patch(
        'karbor.api.v1.restores.RestoresController.'
        '_restore_get')
    def test_restore_show(self, moak_restore_get):
        req = fakes.HTTPRequest.blank('/v1/restores')
        self.controller.show(
            req, '2a9ce1f3-cc1a-4516-9435-0ebb13caa398')
        self.assertTrue(moak_restore_get.called)

    def test_restore_show_Invalid(self):
        req = fakes.HTTPRequest.blank('/v1/restores/1')
        self.assertRaises(
            exc.HTTPBadRequest, self.controller.show,
            req, "1")

    @mock.patch(
        'karbor.api.v1.restores.RestoresController.'
        '_restore_get')
    def test_restore_delete(self, moak_restore_get):
        req = fakes.HTTPRequest.blank('/v1/restores')
        self.controller.show(
            req, '2a9ce1f3-cc1a-4516-9435-0ebb13caa398')
        self.assertTrue(moak_restore_get.called)

    def test_restore_delete_Invalid(self):
        req = fakes.HTTPRequest.blank('/v1/restores/1')
        self.assertRaises(
            exc.HTTPBadRequest, self.controller.show,
            req, "1")

    def _restore_in_request_body(
            self, project_id=DEFAULT_PROJECT_ID,
            provider_id=DEFAULT_PROVIDER_ID,
            checkpoint_id=DEFAULT_CHECKPOINT_ID,
            restore_target=DEFAULT_RESTORE_TARGET,
            restore_auth=DEFAULT_RESTORE_AUTH,
            parameters=DEFAULT_PARAMETERS,
            status=constants.RESOURCE_STATUS_STARTED):
        restore_req = {
            'project_id': project_id,
            'provider_id': provider_id,
            'checkpoint_id': checkpoint_id,
            'restore_target': restore_target,
            'restore_auth': restore_auth,
            'parameters': parameters,
            'status': status,
        }

        return restore_req



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from karbor.common import constants
from karbor.tests.fullstack import karbor_base
from karbor.tests.fullstack import karbor_objects as objects


class PlansTest(karbor_base.KarborBaseTest):
    """Test Plans operation"""
    def setUp(self):
        super(PlansTest, self).setUp()
        self.provider_id = self.provider_id_noop

    def test_plans_list(self):
        # create plan
        volume = self.store(objects.Volume())
        volume.create(1)
        plan1 = self.store(objects.Plan())
        plan1.create(self.provider_id, [volume, ])
        plan2 = self.store(objects.Plan())
        plan2.create(self.provider_id, [volume, ])

        # list plans after creating
        items = self.karbor_client.plans.list()
        ids = [item.id for item in items]
        self.assertTrue(plan1.id in ids)
        self.assertTrue(plan2.id in ids)

    def test_plans_get(self):
        plan_name = "Fullstack Test Get"
        volume = self.store(objects.Volume())
        volume.create(1)
        plan = self.store(objects.Plan())
        plan.create(self.provider_id, [volume, ], name=plan_name)

        plan_item = self.karbor_client.plans.get(plan.id)
        self.assertEqual(plan_name, plan_item.name)

    def test_plans_update(self):
        plan_initial_name = "Fullstack Plan Pre-Update"
        plan_updated_name = "Fullstack Plan Post-Update"
        volume1_name = "Fullstack Plan Update Volume1"
        volume2_name = "Fullstack Plan Update Volume2"
        volume1 = self.store(objects.Volume())
        volume1.create(1, name=volume1_name)
        volume2 = self.store(objects.Volume())
        volume2.create(1, name=volume2_name)
        plan = self.store(objects.Plan())
        plan.create(self.provider_id, [volume1, ], name=plan_initial_name)

        # sanity
        plan_item = self.karbor_client.plans.get(plan.id)
        self.assertEqual(plan_initial_name, plan_item.name)
        self.assertEqual("suspended", plan_item.status)
        self.assertEqual([{"id": volume1.id,
                           "type": constants.VOLUME_RESOURCE_TYPE,
                           "name": volume1_name,
                           "extra_info":
                               {"availability_zone": "az1"}}],
                         plan_item.resources)

        # update name
        data = {"name": plan_updated_name}
        plan_item = self.karbor_client.plans.update(plan.id, data)
        self.assertEqual(plan_updated_name, plan_item.name)

        # update resources
        data = {"resources": [volume2.to_dict(), ]}
        plan_item = self.karbor_client.plans.update(plan.id, data)
        self.assertEqual([{"id": volume2.id,
                           "type": constants.VOLUME_RESOURCE_TYPE,
                           "name": volume2_name,
                           "extra_info":
                               {"availability_zone": "az1"}}],
                         plan_item.resources)

        # update status
        data = {"status": "started"}
        plan_item = self.karbor_client.plans.update(plan.id, data)
        self.assertEqual("started", plan_item.status)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import errno
import os

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from karbor import exception
from karbor.i18n import _
from karbor.services.protection.bank_plugin import BankPlugin

import six

file_system_bank_plugin_opts = [
    cfg.StrOpt('file_system_bank_path',
               help='The file system bank path to use.'),
    cfg.StrOpt('bank_object_container',
               default='karbor',
               help='The file system bank container to use.'),
]

LOG = logging.getLogger(__name__)


class FileSystemBankPlugin(BankPlugin):
    """File system bank plugin"""
    def __init__(self, config):
        super(FileSystemBankPlugin, self).__init__(config)
        self._config.register_opts(file_system_bank_plugin_opts,
                                   "file_system_bank_plugin")
        plugin_cfg = self._config.file_system_bank_plugin
        self.file_system_bank_path = plugin_cfg.file_system_bank_path
        self.bank_object_container = plugin_cfg.bank_object_container

        try:
            self._create_dir(self.file_system_bank_path)
            self.object_container_path = "/".join([self.file_system_bank_path,
                                                   self.bank_object_container])
            self._create_dir(self.object_container_path)
        except OSError as err:
            LOG.exception(_("Init file system bank failed. err: %s"), err)

        self.owner_id = uuidutils.generate_uuid()

    def _validate_path(self, path):
        if path.find('..') >= 0:
            msg = (_("The path(%s) is invalid.") % path)
            raise exception.InvalidInput(msg)

    def _create_dir(self, path):
        try:
            original_umask = None
            try:
                original_umask = os.umask(0)
                os.makedirs(path)
            finally:
                os.umask(original_umask)
        except OSError as err:
            if err.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                LOG.exception(_("Create the directory failed. path: %s"), path)
                raise

    def _write_object(self, path, data):
        obj_file_name = None
        try:
            obj_path = self.object_container_path + path.rsplit('/', 1)[0]
            obj_file_name = self.object_container_path + path
            self._create_dir(obj_path)
            mode = "wb"
            if isinstance(data, six.string_types):
                mode = "w"
            with open(obj_file_name, mode=mode) as obj_file:
                obj_file.write(data)
        except (OSError, IOError):
            LOG.exception(_("Write object failed. name: %s"), obj_file_name)
            raise

    def _get_object(self, path):
        obj_file_name = self.object_container_path + path
        if not os.path.isfile(obj_file_name):
            LOG.exception(_("Object is not a file. name: %s"), obj_file_name)
            raise
        try:
            with open(obj_file_name, mode='r') as obj_file:
                data = obj_file.read()
                return data
        except OSError:
            LOG.exception(_("Get object failed. name: %s"), obj_file_name)
            raise

    def _delete_object(self, path):
        obj_path = self.object_container_path + path.rsplit('/', 1)[0]
        obj_file_name = self.object_container_path + path
        try:
            os.remove(obj_file_name)
            if not os.listdir(obj_path) and (
                    obj_path != self.object_container_path):
                os.rmdir(obj_path)
        except OSError:
            LOG.exception(_("Delete the object failed. name: %s"),
                          obj_file_name)
            raise

    def _list_object(self, path):
        obj_file_path = self.object_container_path + path
        if not os.path.isdir(obj_file_path):
            LOG.exception(_("Path is not a directory. name: %s"),
                          obj_file_path)
            raise
        try:
            if os.path.isdir(obj_file_path):
                return os.listdir(obj_file_path)
            else:
                base_dir = os.path.dirname(obj_file_path)
                base_name = os.path.basename(obj_file_path)
                return (
                    base_dir + '/' + obj_file
                    for obj_file in os.listdir(base_dir)
                    if obj_file.startswith(base_name)
                )
        except OSError:
            LOG.exception(_("List the object failed. path: %s"), obj_file_path)
            raise

    def get_owner_id(self):
        return self.owner_id

    def update_object(self, key, value):
        LOG.debug("FsBank: update_object. key: %s", key)
        self._validate_path(key)
        try:
            if not isinstance(value, str):
                value = jsonutils.dumps(value)
            self._write_object(path=key,
                               data=value)
        except OSError as err:
            LOG.error("Update object failed. err: %s", err)
            raise exception.BankUpdateObjectFailed(reason=err,
                                                   key=key)

    def delete_object(self, key):
        LOG.debug("FsBank: delete_object. key: %s", key)
        self._validate_path(key)
        try:
            self._delete_object(path=key)
        except OSError as err:
            LOG.error("Delete object failed. err: %s", err)
            raise exception.BankDeleteObjectFailed(reason=err,
                                                   key=key)

    def get_object(self, key):
        LOG.debug("FsBank: get_object. key: %s", key)
        self._validate_path(key)
        try:
            data = self._get_object(path=key)
        except OSError as err:
            LOG.error("Get object failed. err: %s", err)
            raise exception.BankGetObjectFailed(reason=err,
                                                key=key)
        if isinstance(data, six.string_types):
            try:
                data = jsonutils.loads(data)
            except ValueError:
                pass
        return data

    def list_objects(self, prefix=None, limit=None, marker=None,
                     sort_dir=None):
        LOG.debug("FsBank: list_objects. key: %s", prefix)
        try:
            file_lists = self._list_object(prefix)
            return (
                file_lists[-limit:] if limit is not None else file_lists)
        except OSError as err:
            LOG.error("List objects failed. err: %s", err)
            raise exception.BankListObjectsFailed(reason=err)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import tempfile

from oslo_config import cfg
from oslo_config import fixture
from oslo_utils import importutils

from karbor import exception
from karbor.tests import base


class FileSystemBankPluginTest(base.TestCase):
    def setUp(self):
        super(FileSystemBankPluginTest, self).setUp()

        import_str = (
            "karbor.services.protection.bank_plugins."
            "file_system_bank_plugin.FileSystemBankPlugin"
        )
        plugin_config = cfg.ConfigOpts()
        plugin_config_fixture = self.useFixture(fixture.Config(plugin_config))
        plugin_config_fixture.load_raw_values(
            group='file_system_bank_plugin',
            file_system_bank_path=tempfile.mkdtemp(),
        )
        fs_bank_plugin_cls = importutils.import_class(
            import_str=import_str)
        self.fs_bank_plugin = fs_bank_plugin_cls(plugin_config)

    def test_delete_object(self):
        self.fs_bank_plugin.update_object("/key", "value")
        self.fs_bank_plugin.delete_object("/key")
        object_file = (
            self.fs_bank_plugin.object_container_path + "/key")
        self.assertEqual(False, os.path.isfile(object_file))

    def test_get_object(self):
        self.fs_bank_plugin.update_object("/key", "value")
        value = self.fs_bank_plugin.get_object("/key")
        self.assertEqual("value", value)

    def test_list_objects(self):
        self.fs_bank_plugin.update_object("/list/key-1", "value-1")
        self.fs_bank_plugin.update_object("/list/key-2", "value-2")
        objects = self.fs_bank_plugin.list_objects(prefix="/list")
        self.assertEqual(2, len(objects))
        self.assertIn('key-1', objects)
        self.assertIn('key-2', objects)

    def test_update_object(self):
        self.fs_bank_plugin.update_object("/key-1", "value-1")
        self.fs_bank_plugin.update_object("/key-1", "value-2")
        object_file = (
            self.fs_bank_plugin.object_container_path + "/key-1")
        with open(object_file, "r") as f:
            contents = f.read()
        self.assertEqual("value-2", contents)

    def test_update_object_with_invaild_path(self):
        self.assertRaises(exception.InvalidInput,
                          self.fs_bank_plugin.update_object,
                          "../../../../../../../etc/shadow",
                          "value-1")

    def test_create_get_dict_object(self):
        self.fs_bank_plugin.update_object("/index.json",
                                          {"key": "value"})
        value = self.fs_bank_plugin.get_object(
            "/index.json")
        self.assertEqual({"key": "value"}, value)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import namedtuple
import mock
from oslo_config import cfg

from karbor.common import constants
from karbor.context import RequestContext
from karbor.resource import Resource
from karbor.services.protection.bank_plugin import Bank
from karbor.services.protection.bank_plugin import BankPlugin
from karbor.services.protection.bank_plugin import BankSection
from karbor.services.protection.protection_plugins.server \
    import server_plugin_schemas
from karbor.services.protection.protection_plugins.server. \
    nova_protection_plugin import NovaProtectionPlugin
from karbor.tests import base


class Server(object):
    def __init__(self, id, addresses, availability_zone,
                 flavor, key_name, security_groups):
        super(Server, self).__init__()
        self.id = id
        self.addresses = addresses
        self.__setattr__("OS-EXT-AZ:availability_zone", availability_zone)
        self.flavor = flavor
        self.key_name = key_name
        self.security_groups = security_groups


class Volume(object):
    def __init__(self, id, volume_type, status, bootable,
                 attachments, name=None):
        super(Volume, self).__init__()
        self.id = id
        self.volume_type = volume_type
        self.status = status
        self.bootable = bootable
        self.attachments = attachments
        self.name = name


class Image(object):
    def __init__(self, id, status, disk_format, container_format):
        super(Image, self).__init__()
        self.id = id
        self.status = status
        self.disk_format = disk_format
        self.container_format = container_format


FakePorts = {'ports': [
    {'fixed_ips': [{'subnet_id': 'subnet-1',
                    'ip_address': '10.0.0.21'}],
     'id': 'port-1',
     'mac_address': 'mac_address_1',
     'device_id': 'vm_id_1',
     'name': '',
     'admin_state_up': True,
     'network_id': 'network_id_1'},
    {'fixed_ips': [{'subnet_id': 'subnet-1',
                    'ip_address': '10.0.0.22'}],
     'id': 'port-2',
     'mac_address': 'mac_address_2',
     'device_id': 'vm_id_2',
     'name': '',
     'admin_state_up': True,
     'network_id': 'network_id_2'}
]}

FakeServers = {
    "vm_id_1": Server(id="vm_id_1",
                      addresses={'fake_net': [
                          {'OS-EXT-IPS-MAC:mac_addr': 'mac_address_1',
                           'OS-EXT-IPS:type': 'fixed',
                           'addr': '10.0.0.21',
                           'version': 4}
                      ]},
                      availability_zone="nova",
                      flavor={'id': 'flavor_id',
                              'links': [
                                  {'href': '',
                                   'rel': 'bookmark'}
                              ]},
                      key_name=None,
                      security_groups="default"),
    "vm_id_2": Server(id="vm_id_2",
                      addresses={'fake_net': [
                          {'OS-EXT-IPS-MAC:mac_addr': 'mac_address_2',
                           'OS-EXT-IPS:type': 'fixed',
                           'addr': '10.0.0.22',
                           'version': 4}
                      ]},
                      availability_zone="nova",
                      flavor={'id': 'flavor_id',
                              'links': [
                                  {'href': '',
                                   'rel': 'bookmark'}
                              ]},
                      key_name=None,
                      security_groups="default")
}

FakeVolumes = {
    "vol_id_1": Volume(id="vol_id_1",
                       volume_type="",
                       status="in-use",
                       bootable="",
                       attachments=[{'server_id': 'vm_id_2',
                                     'attachment_id': '',
                                     'host_name': '',
                                     'volume_id': 'vol_id_1',
                                     'device': '/dev/vdb',
                                     'id': 'attach_id_1'}],
                       name="vol_id_1_name")
}

FakeImages = {
    "image_id_1": Image(id="image_id_1",
                        disk_format="",
                        container_format="",
                        status="active")
}

FakeGraphNode = namedtuple("GraphNode", (
    "value",
    "child_nodes",
))


class FakeNovaClient(object):
    class Servers(object):
        def get(self, server_id):
            return FakeServers[server_id]

        def create_image(self, server_id, name, **kwargs):
            FakeImages["image_id_2"] = Image(id="image_id_2",
                                             disk_format="",
                                             container_format="",
                                             status="active")
            return "image_id_2"

        def __getattr__(self, item):
            return None

    def __init__(self):
        super(FakeNovaClient, self).__init__()
        self.servers = self.Servers()


class FakeGlanceClient(object):
    class Images(object):
        def get(self, image_id):
            return FakeImages[image_id]

        def data(self, image_id):
            return "image_data_" + image_id

        def delete(self, image_id):
            if image_id in FakeImages:
                FakeImages.pop(image_id)

        def __getattr__(self, item):
            return None

    def __init__(self):
        super(FakeGlanceClient, self).__init__()
        self.images = self.Images()


class FakeCinderClient(object):
    class Volumes(object):
        def get(self, volume_id):
            return FakeVolumes[volume_id]

        def list(self, detailed=True, search_opts=None, limit=None):
            return [FakeVolumes['vol_id_1'], ]

        def __getattr__(self, item):
            return None

    def __init__(self):
        super(FakeCinderClient, self).__init__()
        self.volumes = self.Volumes()


class FakeNeutronClient(object):
    def list_ports(self, mac_address):
        result_ports = []
        for port in FakePorts["ports"]:
            if port["mac_address"] == mac_address:
                result_ports.append(port)
        return {"ports": result_ports}


class FakeBankPlugin(BankPlugin):
    def __init__(self, config=None):
        super(FakeBankPlugin, self).__init__(config)
        self._objects = {}

    def update_object(self, key, value):
        self._objects[key] = value

    def get_object(self, key):
        value = self._objects.get(key, None)
        if value is None:
            raise Exception
        return value

    def list_objects(self, prefix=None, limit=None, marker=None,
                     sort_dir=None):
        objects_name = []
        if prefix is not None:
            for key, value in self._objects.items():
                if key.find(prefix) == 0:
                    objects_name.append(key)
        else:
            objects_name = self._objects.keys()
        return objects_name

    def delete_object(self, key):
        self._objects.pop(key)

    def get_owner_id(self):
        return


fake_bank = Bank(FakeBankPlugin())

ResourceNode = namedtuple(
    "ResourceNode",
    ["value",
     "child_nodes"]
)


class Checkpoint(object):
    def __init__(self):
        super(Checkpoint, self).__init__()
        self.id = "checkpoint_id"
        self.graph = []

    @property
    def resource_graph(self):
        return self.graph

    @resource_graph.setter
    def resource_graph(self, resource_graph):
        self.graph = resource_graph

    def get_resource_bank_section(self, resource_id):
        return BankSection(
            bank=fake_bank,
            section="/resource_data/%s/%s" % (self.id, resource_id)
        )


def call_hooks(operation, checkpoint, resource, context, parameters, **kwargs):
    def noop(*args, **kwargs):
        pass

    hooks = (
        'on_prepare_begin',
        'on_prepare_finish',
        'on_main',
        'on_complete',
    )
    for hook_name in hooks:
        hook = getattr(operation, hook_name, noop)
        hook(checkpoint, resource, context, parameters, **kwargs)


class NovaProtectionPluginTest(base.TestCase):
    def setUp(self):
        super(NovaProtectionPluginTest, self).setUp()
        self.cntxt = RequestContext(user_id='demo',
                                    project_id='abcd',
                                    auth_token='efgh')
        self.plugin = NovaProtectionPlugin(cfg.CONF)
        self.glance_client = FakeGlanceClient()
        self.nova_client = FakeNovaClient()
        self.cinder_client = FakeCinderClient()
        self.neutron_client = FakeNeutronClient()
        self.checkpoint = Checkpoint()

    def test_get_options_schema(self):
        options_schema = self.plugin.get_options_schema(
            constants.SERVER_RESOURCE_TYPE)
        self.assertEqual(options_schema, server_plugin_schemas.OPTIONS_SCHEMA)

    def test_get_restore_schema(self):
        options_schema = self.plugin.get_restore_schema(
            constants.SERVER_RESOURCE_TYPE)
        self.assertEqual(options_schema, server_plugin_schemas.RESTORE_SCHEMA)

    def test_get_saved_info_schema(self):
        options_schema = self.plugin.get_saved_info_schema(
            constants.SERVER_RESOURCE_TYPE)
        self.assertEqual(options_schema,
                         server_plugin_schemas.SAVED_INFO_SCHEMA)

    @mock.patch('karbor.services.protection.clients.neutron.create')
    @mock.patch('karbor.services.protection.clients.glance.create')
    @mock.patch('karbor.services.protection.clients.nova.create')
    @mock.patch('karbor.services.protection.clients.cinder.create')
    def test_create_backup_without_volumes(self, mock_cinder_client,
                                           mock_nova_client,
                                           mock_glance_client,
                                           mock_neutron_client):
        resource = Resource(id="vm_id_1",
                            type=constants.SERVER_RESOURCE_TYPE,
                            name="fake_vm")

        protect_operation = self.plugin.get_protect_operation(resource)
        mock_cinder_client.return_value = self.cinder_client
        mock_nova_client.return_value = self.nova_client
        mock_glance_client.return_value = self.glance_client
        mock_neutron_client.return_value = self.neutron_client

        call_hooks(protect_operation, self.checkpoint, resource, self.cntxt,
                   {})

        self.assertEqual(
            constants.RESOURCE_STATUS_AVAILABLE,
            fake_bank._plugin._objects[
                "/resource_data/checkpoint_id/vm_id_1/status"]
        )
        resource_definition = {
            "resource_id": "vm_id_1",
            "attach_metadata": {},
            'boot_metadata': {'boot_device_type': 'volume'},
            "server_metadata": {
                "availability_zone": "nova",
                "networks": ["network_id_1"],
                "floating_ips": [],
                "flavor": "flavor_id",
                "key_name": None,
                "security_groups": "default",
            },
        }
        self.assertEqual(
            resource_definition,
            fake_bank._plugin._objects[
                "/resource_data/checkpoint_id/vm_id_1/metadata"]
        )

    @mock.patch('karbor.services.protection.clients.neutron.create')
    @mock.patch('karbor.services.protection.clients.glance.create')
    @mock.patch('karbor.services.protection.clients.nova.create')
    @mock.patch('karbor.services.protection.clients.cinder.create')
    def test_create_backup_with_volumes(self, mock_cinder_client,
                                        mock_nova_client,
                                        mock_glance_client,
                                        mock_neutron_client):
        vm_resource = Resource(id="vm_id_2",
                               type=constants.SERVER_RESOURCE_TYPE,
                               name="fake_vm")

        protect_operation = self.plugin.get_protect_operation(vm_resource)
        mock_cinder_client.return_value = self.cinder_client
        mock_nova_client.return_value = self.nova_client
        mock_glance_client.return_value = self.glance_client
        mock_neutron_client.return_value = self.neutron_client
        checkpoint = Checkpoint()
        checkpoint.resource_graph = [FakeGraphNode(value=Resource(
            type='OS::Nova::Server', id='vm_id_2', name='None'),
            child_nodes=[FakeGraphNode(value=Resource(
                type='OS::Cinder::Volume', id='vol_id_1', name=None),
                child_nodes=())])]

        call_hooks(protect_operation, checkpoint, vm_resource, self.cntxt,
                   {})

        self.assertEqual(
            constants.RESOURCE_STATUS_AVAILABLE,
            fake_bank._plugin._objects[
                "/resource_data/checkpoint_id/vm_id_2/status"]
        )
        resource_definition = {
            "resource_id": "vm_id_2",
            "attach_metadata": {"vol_id_1": "/dev/vdb"},
            'boot_metadata': {'boot_device_type': 'volume'},
            "server_metadata": {
                "availability_zone": "nova",
                "networks": ["network_id_2"],
                "floating_ips": [],
                "flavor": "flavor_id",
                "key_name": None,
                "security_groups": "default",
            },
            'attach_metadata': {
                'vol_id_1': {'attachment_id': '',
                             'bootable': '',
                             'device': '/dev/vdb',
                             'host_name': '',
                             'id': 'attach_id_1',
                             'server_id': 'vm_id_2',
                             'volume_id': 'vol_id_1'}
            },
        }
        self.assertEqual(
            fake_bank._plugin._objects[
                "/resource_data/checkpoint_id/vm_id_2/metadata"],
            resource_definition
        )

    @mock.patch('karbor.services.protection.clients.glance.create')
    def test_delete_backup(self, mock_glance_client):
        resource = Resource(id="vm_id_1",
                            type=constants.SERVER_RESOURCE_TYPE,
                            name="fake_vm")

        fake_bank._plugin._objects[
            "/resource_data/checkpoint_id/vm_id_1/metadata"] = {
            "resource_id": "vm_id_1",
            "attach_metadata": {},
            "server_metadata": {
                "availability_zone": "nova",
                "networks": ["network_id_1"],
                "floating_ips": [],
                "flavor": "flavor_id",
                "key_name": None,
                "security_groups": "default",
            },
            "snapshot_id": "image_id_2",
            "snapshot_metadata": {
                "disk_format": "",
                "container_format": "",
                "name": "snapshot_checkpoint_id@vm_id_1"
            }
        }

        delete_operation = self.plugin.get_delete_operation(resource)
        mock_glance_client.return_value = self.glance_client

        fake_bank._plugin._objects[
            "/resource_data/checkpoint_id/vm_id_1/data_0"
        ] = "image_data_1"
        fake_bank._plugin._objects[
            "/resource_data/checkpoint_id/vm_id_1/data_1"
        ] = "image_data_1"

        call_hooks(delete_operation, self.checkpoint, resource, self.cntxt,
                   {})

    def test_get_supported_resources_types(self):
        types = self.plugin.get_supported_resources_types()
        self.assertEqual([constants.SERVER_RESOURCE_TYPE], types)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_messaging.rpc import dispatcher as rpc_dispatcher

from karbor.common import constants
from karbor import context
from karbor import exception
from karbor import objects
from karbor.services.operationengine import manager as service_manager
from karbor.tests import base


class FakeTriggerManager(object):

    def __init__(self):
        super(FakeTriggerManager, self).__init__()
        self._trigger = {}

    def register_operation(self, trigger_id, operation_id, **kwargs):
        if trigger_id not in self._trigger:
            self._trigger[trigger_id] = []

        if operation_id in self._trigger[trigger_id]:
            raise exception.ScheduledOperationExist(op_id=operation_id)

        self._trigger[trigger_id].append(operation_id)

    def unregister_operation(self, trigger_id, operation_id, **kwargs):
        pass

    def add_trigger(self, trigger_id, trigger_type, trigger_property):
        self._trigger[trigger_id] = []


class FakeUserTrustManager(object):
    def add_operation(self, context, operation_id):
        return "123"

    def delete_operation(self, context, operation_id):
        pass

    def resume_operation(self, operation_id, user_id, project_id, trust_id):
        pass


class OperationEngineManagerTestCase(base.TestCase):
    """Test cases for OperationEngineManager class."""

    def setUp(self):
        super(OperationEngineManagerTestCase, self).setUp()

        self.manager = service_manager.OperationEngineManager()
        self.manager._service_id = 0
        self.manager._trigger_manager = FakeTriggerManager()
        self.manager._user_trust_manager = FakeUserTrustManager()

        self.ctxt = context.get_admin_context()
        self._trigger = self._create_one_trigger()
        self._operation = self._create_scheduled_operation(self._trigger.id)

    def test_init_host(self):
        trigger_id = self._trigger.id
        operation_id = self._operation.id

        self._create_operation_state(operation_id)

        op = self._create_scheduled_operation(self._trigger.id, False)
        self._create_operation_state(op.id)

        self.manager._restore()

        trigger_manager = self.manager._trigger_manager
        self.assertIn(trigger_id, trigger_manager._trigger)
        self.assertIn(operation_id, trigger_manager._trigger[trigger_id])
        self.assertNotIn(op.id, trigger_manager._trigger[trigger_id])

    def test_create_operation(self):
        op = self._create_scheduled_operation(self._trigger.id, False)
        with mock.patch(
            'karbor.services.operationengine.operations.protect_operation.'
            'ProtectOperation.check_operation_definition'
        ):
            self.manager.create_scheduled_operation(
                self.ctxt, op)

        state_obj = objects.ScheduledOperationState.get_by_operation_id(
            self.ctxt, op.id)

        self.assertIsNotNone(state_obj)

    def test_create_operation_invalid_operation_definition(self):
        op = self._create_scheduled_operation(self._trigger.id, False)
        self.assertRaises(
            rpc_dispatcher.ExpectedException,
            self.manager.create_scheduled_operation,
            self.ctxt,
            op,
        )

    def test_create_operation_invalid_operation_type(self):
        op = self._create_scheduled_operation(self._trigger.id, False)
        op.operation_type = "123"
        self.assertRaises(
            rpc_dispatcher.ExpectedException,
            self.manager.create_scheduled_operation,
            self.ctxt,
            op,
        )

    def test_delete_operation_get_state_failed(self):
        self.assertRaises(rpc_dispatcher.ExpectedException,
                          self.manager.delete_scheduled_operation,
                          self.ctxt, self._operation.id, 1)

    def test_delete_operation(self):
        state = self._create_operation_state(self._operation.id)

        self.manager.delete_scheduled_operation(
            self.ctxt, self._operation.id, 1)

        state = objects.ScheduledOperationState.get_by_operation_id(
            self.ctxt, self._operation.id)
        self.assertEqual(constants.OPERATION_STATE_DELETED, state.state)

    @mock.patch.object(FakeTriggerManager, 'unregister_operation')
    def test_suspend_resume_operation(self, unregister):
        op_id = 'suspend'
        trigger_id = "trigger"

        self.manager.resume_scheduled_operation(self.ctxt, op_id, trigger_id)
        self.assertIn(op_id,
                      self.manager._trigger_manager._trigger[trigger_id])

        self.manager.resume_scheduled_operation(self.ctxt, op_id, trigger_id)
        self.assertEqual(1, len(
            self.manager._trigger_manager._trigger[trigger_id]))

        # resume
        self.manager.suspend_scheduled_operation(self.ctxt, op_id, trigger_id)
        unregister.assert_called_once_with(trigger_id, op_id)

    def _create_one_trigger(self):
        trigger_info = {
            'project_id': "123",
            "name": "123",
            "type": "time",
            "properties": {
                "format": "crontab",
                "pattern": "* * * * *"
            },
        }
        trigger = objects.Trigger(self.ctxt, **trigger_info)
        trigger.create()
        return trigger

    def _create_scheduled_operation(self, trigger_id, enabled=True):
        operation_info = {
            "name": "123",
            'description': '123',
            "operation_type": "protect",
            'user_id': '123',
            "project_id": "123",
            "trigger_id": trigger_id,
            "operation_definition": {
                "plan_id": ""
            },
            "enabled": enabled
        }
        operation = objects.ScheduledOperation(self.ctxt, **operation_info)
        operation.create()
        return operation

    def _create_operation_state(self, operation_id):
        state_info = {
            "operation_id": operation_id,
            "service_id": self.manager._service_id,
            'trust_id': '123',
            "state": constants.OPERATION_STATE_REGISTERED
        }
        operation_state = objects.ScheduledOperationState(context,
                                                          **state_info)
        operation_state.create()
        return operation_state



# Copyright 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import futurist
import mock

from oslo_config import cfg
from oslo_log import log as logging

from karbor.resource import Resource
from karbor.services.protection.bank_plugin import Bank
from karbor.services.protection.bank_plugin import BankPlugin
from karbor.services.protection.bank_plugin import BankSection
from karbor.services.protection.graph import build_graph
from karbor.services.protection import protection_plugin
from karbor.services.protection import provider
from karbor.services.protection import resource_flow

from taskflow import engines
from taskflow.patterns import graph_flow
from taskflow.patterns import linear_flow
from taskflow import task

LOG = logging.getLogger(__name__)

A = Resource(id='A', type='fake', name='fake')
B = Resource(id='B', type='fake', name='fake')
C = Resource(id='C', type='fake', name='fake')
D = Resource(id='D', type='fake', name='fake')
E = Resource(id='E', type='fake', name='fake')


resource_map = {
    A: [C],
    B: [C],
    C: [D, E],
    D: [],
    E: [],
}

resource_graph = build_graph([A, B, C, D], resource_map.__getitem__)


def fake_protection_plan():
    protection_plan = {'id': 'fake_id',
                       'is_enabled': True,
                       'name': 'fake_protection_plan',
                       'comments': '',
                       'revision': 0,
                       'resources': [
                           {"id": "A", "type": "fake", "name": "fake"},
                           {"id": "B", "type": "fake", "name": "fake"},
                           {"id": "C", "type": "fake", "name": "fake"},
                           {"id": "D", "type": "fake", "name": "fake"}],
                       'protection_provider': None,
                       'parameters': {},
                       'provider_id': 'fake_id'
                       }
    return protection_plan


plan_resources = [A, B, C, D]


class FakeBankPlugin(BankPlugin):
    def __init__(self, config=None):
        super(FakeBankPlugin, self).__init__(config=config)
        self._objects = {}
        fake_bank_opts = [
            cfg.HostAddressOpt('fake_host'),
        ]
        if config:
            config.register_opts(fake_bank_opts, 'fake_bank')
            self.fake_host = config['fake_bank']['fake_host']

    def update_object(self, key, value):
        self._objects[key] = value

    def get_object(self, key):
        value = self._objects.get(key, None)
        if value is None:
            raise Exception
        return value

    def list_objects(self, prefix=None, limit=None, marker=None):
        objects_name = []
        if prefix is not None:
            for key, value in self._objects.items():
                if key.find(prefix) == 0:
                    objects_name.append(key.lstrip(prefix))
        else:
            objects_name = self._objects.keys()
        return objects_name

    def delete_object(self, key):
        self._objects.pop(key)

    def get_owner_id(self):
        return


def fake_restore():
    restore = {
        'id': 'fake_id',
        'provider_id': 'fake_provider_id',
        'checkpoint_id': 'fake_checkpoint_id',
        'parameters': {
            'username': 'fake_username',
            'password': 'fake_password'
        },
        'restore_target': 'fake_target_url',
    }
    return restore


class FakeProtectablePlugin(object):
    def get_resource_type(self):
        pass

    def get_parent_resource_types(self):
        pass

    def list_resources(self):
        pass

    def get_dependent_resources(self, parent_resource):
        pass


class MockOperation(protection_plugin.Operation):
    def __init__(self):
        super(MockOperation, self).__init__()
        for hook_name in resource_flow.HOOKS:
            setattr(self, hook_name, mock.Mock())


class FakeOperation(protection_plugin.Operation):
    def __init__(self):
        super(FakeOperation, self).__init__()
        self.all_invokes = {}

    def _update_invokes(self, resource, func, info):
        self.all_invokes.setdefault(resource, {})[func] = info

    def on_prepare_begin(self, checkpoint, resource, context, parameters,
                         **kwargs):
        info = {
            'checkpoint': checkpoint,
            'resource': resource,
            'context': context,
            'parameters': parameters,
            'kwargs': kwargs
        }
        self._update_invokes(resource, 'on_prepare_begin', info)

    def on_prepare_finish(self, checkpoint, resource, context, parameters,
                          **kwargs):
        info = {
            'checkpoint': checkpoint,
            'resource': resource,
            'context': context,
            'parameters': parameters,
            'kwargs': kwargs
        }
        self._update_invokes(resource, 'on_prepare_finish', info)

    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        info = {
            'checkpoint': checkpoint,
            'resource': resource,
            'context': context,
            'parameters': parameters,
            'kwargs': kwargs
        }
        self._update_invokes(resource, 'on_main', info)

    def on_complete(self, checkpoint, resource, context, parameters, **kwargs):
        info = {
            'checkpoint': checkpoint,
            'resource': resource,
            'context': context,
            'parameters': parameters,
            'kwargs': kwargs
        }
        self._update_invokes(resource, 'on_complete', info)


class FakeProtectionPlugin(protection_plugin.ProtectionPlugin):
    SUPPORTED_RESOURCES = [
        'Test::ResourceA',
        'Test::ResourceB',
        'Test::ResourceC',
    ]

    def __init__(self, config=None, *args, **kwargs):
        super(FakeProtectionPlugin, self).__init__(config)
        fake_plugin_opts = [
            cfg.StrOpt('fake_user'),
        ]
        if config:
            config.register_opts(fake_plugin_opts, 'fake_plugin')
            self.fake_user = config['fake_plugin']['fake_user']

    def get_protect_operation(self, *args, **kwargs):
        return MockOperation()

    def get_restore_operation(self, *args, **kwargs):
        return MockOperation()

    def get_delete_operation(self, *args, **kwargs):
        return MockOperation()

    @classmethod
    def get_supported_resources_types(cls):
        return cls.SUPPORTED_RESOURCES

    @classmethod
    def get_options_schema(cls, resource_type):
        return {}

    @classmethod
    def get_saved_info_schema(cls, resource_type):
        return {}

    @classmethod
    def get_restore_schema(cls, resource_type):
        return {}

    @classmethod
    def get_saved_info(cls, metadata_store, resource):
        return None


class FakeCheckpoint(object):
    def __init__(self):
        super(FakeCheckpoint, self).__init__()
        self.id = 'fake_checkpoint'
        self.status = 'available'
        self.resource_graph = resource_graph

    def purge(self):
        pass

    def commit(self):
        pass

    def get_resource_bank_section(self, resource_id):
        bank = Bank(FakeBankPlugin())
        return BankSection(bank, resource_id)

    def to_dict(self):
        return {
            "id": self.id,
            "status": self.status,
            "resource_graph": self.resource_graph,
            "protection_plan": None,
            "project_id": None
        }


class FakeCheckpointCollection(object):
    def create(self, plan, checkpoint_properties=None):
        return FakeCheckpoint()

    def get(self, checkpoint_id):
        return FakeCheckpoint()


class FakeProvider(provider.PluggableProtectionProvider):
    def __init__(self):
        self._id = 'test'
        self._name = 'provider'
        self._description = 'fake_provider'
        self._extend_info_schema = {}
        self._config = None
        self._plugin_map = {
            'fake': FakeProtectionPlugin,
        }

    def get_checkpoint_collection(self):
        return FakeCheckpointCollection()


class FakeFlowEngine(object):
    def create_task(self, function, requires=None, provides=None,
                    inject=None, **kwargs):
        name = kwargs.get('name', None)
        auto_extract = kwargs.get('auto_extract', True)
        rebind = kwargs.get('rebind', None)
        revert = kwargs.get('revert', None)
        version = kwargs.get('version', None)
        if function:
            return task.FunctorTask(function,
                                    name=name,
                                    provides=provides,
                                    requires=requires,
                                    auto_extract=auto_extract,
                                    rebind=rebind,
                                    revert=revert,
                                    version=version,
                                    inject=inject)

    def add_tasks(self, flow, *nodes, **kwargs):
        if flow is None:
            LOG.error("The flow is None, get it first")
        flow.add(*nodes, **kwargs)

    def link_task(self, flow, u, v):
        flow.link(u, v)

    def build_flow(self, flow_name, flow_type='graph'):
        if flow_type == 'linear':
            return linear_flow.Flow(flow_name)
        elif flow_type == 'graph':
            return graph_flow.Flow(flow_name)
        else:
            LOG.error("unsupported flow type:%s", flow_type)
            return

    def get_engine(self, flow, **kwargs):
        if flow is None:
            LOG.error("Flow is None, build it first")
            return
        executor = kwargs.get('executor', None)
        engine = kwargs.get('engine', None)
        store = kwargs.get('store', None)
        if not executor:
            executor = futurist.GreenThreadPoolExecutor()
        if not engine:
            engine = 'parallel'
        flow_engine = engines.load(flow,
                                   executor=executor,
                                   engine=engine,
                                   store=store)
        return flow_engine

    def run_engine(self, flow_engine):
        if flow_engine is None:
            LOG.error("Flow engine is None,get it first")
            return
        flow_engine.run()

    def output(self, flow_engine, target=None):
        if flow_engine is None:
            LOG.error("Flow engine is None,return nothing")
            return
        if target:
            return flow_engine.storage.fetch(target)
        return flow_engine.storage.fetch_all()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from karbor import db
from karbor import exception
from karbor.i18n import _
from karbor import objects
from karbor.objects import base


@base.KarborObjectRegistry.register
class OperationLog(base.KarborPersistentObject, base.KarborObject,
                   base.KarborObjectDictCompat,
                   base.KarborComparableObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(),
        'project_id': fields.UUIDField(),
        'scheduled_operation_id': fields.UUIDField(),
        'started_at': base.DateTimeField(nullable=True),
        'ended_at': base.DateTimeField(nullable=True),
        'state': fields.StringField(nullable=True),
        'error': fields.StringField(nullable=True),
        'entries': fields.StringField(nullable=True),
    }

    @staticmethod
    def _from_db_object(context, operation_log, db_operation_log):
        for name, field in operation_log.fields.items():
            value = db_operation_log.get(name)
            if isinstance(field, fields.IntegerField):
                value = value or 0
            elif isinstance(field, fields.DateTimeField):
                value = value or None
            operation_log[name] = value

        operation_log._context = context
        operation_log.obj_reset_changes()
        return operation_log

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already created'))
        updates = self.karbor_obj_get_changes()
        db_operation_log = db.operation_log_create(self._context, updates)
        self._from_db_object(self._context, self, db_operation_log)

    @base.remotable
    def save(self):
        updates = self.karbor_obj_get_changes()
        if updates:
            db.operation_log_update(self._context, self.id, updates)
            self.obj_reset_changes()

    @base.remotable
    def destroy(self):
        with self.obj_as_admin():
            db.operation_log_destroy(self._context, self.id)


@base.KarborObjectRegistry.register
class OperationLogList(base.ObjectListBase, base.KarborObject):
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('OperationLog'),
    }

    @base.remotable_classmethod
    def get_all(cls, context, marker, limit, sort_keys=None, sort_dirs=None,
                filters=None, offset=None):
        operation_logs = db.operation_log_get_all(context, marker, limit,
                                                  sort_keys=sort_keys,
                                                  sort_dirs=sort_dirs,
                                                  filters=filters,
                                                  offset=offset)
        return base.obj_make_list(context, cls(context), objects.OperationLog,
                                  operation_logs)

    @base.remotable_classmethod
    def get_all_by_project(cls, context, project_id, marker, limit,
                           sort_keys=None, sort_dirs=None, filters=None,
                           offset=None):
        operation_logs = db.operation_log_get_all_by_project(
            context, project_id, marker, limit, sort_keys=sort_keys,
            sort_dirs=sort_dirs, filters=filters, offset=offset)
        return base.obj_make_list(context, cls(context), objects.OperationLog,
                                  operation_logs)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

#  operation type
OPERATION_TYPES = (
    OPERATION_PROTECT,
    OPERATION_RESTORE,
    OPERATION_DELETE,
    OPERATION_START,
    OPERATION_SUSPEND,
) = (
    'protect',
    'restore',
    'delete',
    'start',
    'suspend',
)


#  plugin type
PLUGIN_BANK = 'bank'

# supported network resource types
NETWORK_RESOURCE_TYPES = (NET_RESOURCE_TYPE,
                          SUBNET_RESOURCE_TYPE,
                          ROUTER_RESOURCE_TYPE,
                          ROUTERINTERFACE_RESOURCE_TYPE,
                          PORT_RESOURCE_TYPE,
                          SECURITYGROUP_RESOURCE_TYPE,
                          ) = ('OS::Neutron::Net',
                               'OS::Neutron::Subnet',
                               'OS::Neutron::Router',
                               'OS::Neutron::RouterInterface',
                               'OS::Neutron::Port',
                               'OS::Neutron::SecurityGroup',
                               )

# supported resource types
RESOURCE_TYPES = (PROJECT_RESOURCE_TYPE,
                  SERVER_RESOURCE_TYPE,
                  VOLUME_RESOURCE_TYPE,
                  IMAGE_RESOURCE_TYPE,
                  SHARE_RESOURCE_TYPE,
                  NETWORK_RESOURCE_TYPE,
                  ) = ('OS::Keystone::Project',
                       'OS::Nova::Server',
                       'OS::Cinder::Volume',
                       'OS::Glance::Image',
                       'OS::Manila::Share',
                       'OS::Neutron::Network',
                       )
# plan status
PLAN_STATUS_SUSPENDED = 'suspended'
PLAN_STATUS_STARTED = 'started'

CHECKPOINT_STATUS_ERROR = 'error'
CHECKPOINT_STATUS_PROTECTING = 'protecting'
CHECKPOINT_STATUS_AVAILABLE = 'available'
CHECKPOINT_STATUS_DELETING = 'deleting'
CHECKPOINT_STATUS_DELETED = 'deleted'
CHECKPOINT_STATUS_ERROR_DELETING = 'error-deleting'

# resource status
RESOURCE_STATUS_ERROR = 'error'
RESOURCE_STATUS_PROTECTING = 'protecting'
RESOURCE_STATUS_STARTED = 'started'
RESOURCE_STATUS_AVAILABLE = 'available'
RESOURCE_STATUS_DELETING = 'deleting'
RESOURCE_STATUS_DELETED = 'deleted'
RESOURCE_STATUS_RESTORING = 'restoring'  # use in restore object only
RESOURCE_STATUS_UNDEFINED = 'undefined'

# scheduled operation state
OPERATION_STATE_INIT = 'init'
OPERATION_STATE_REGISTERED = 'registered'
OPERATION_STATE_RUNNING = 'running'
OPERATION_STATE_DELETED = 'deleted'

# scheduled operation run type
OPERATION_RUN_TYPE_EXECUTE = 'execute'
OPERATION_RUN_TYPE_RESUME = 'resume'

# scheduled operation execution state
OPERATION_EXE_STATE_IN_PROGRESS = 'in_progress'
OPERATION_EXE_STATE_SUCCESS = 'success'
OPERATION_EXE_STATE_FAILED = 'failed'
OPERATION_EXE_STATE_DROPPED_OUT_OF_WINDOW = 'dropped_out_of_window'

RESTORE_STATUS_SUCCESS = 'success'
RESTORE_STATUS_FAILURE = 'fail'
RESTORE_STATUS_IN_PROGRESS = 'in_progress'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from karbor.services.protection.clients import s3
from karbor.tests import base
from karbor.tests.unit.protection.fake_s3_client import FakeS3Client
import math
import mock
from oslo_config import cfg
from oslo_utils import importutils
import time

CONF = cfg.CONF


class FakeConf(object):
    def __init__(self):
        super(FakeConf, self).__init__()
        self.lease_expire_window = 600
        self.lease_renew_window = 120
        self.lease_validity_window = 100


class S3BankPluginTest(base.TestCase):
    def setUp(self):
        super(S3BankPluginTest, self).setUp()
        self.conf = FakeConf()
        self.fake_connection = FakeS3Client.connection()
        s3.create = mock.MagicMock()
        s3.create.return_value = self.fake_connection
        import_str = (
            "karbor.services.protection.bank_plugins."
            "s3_bank_plugin.S3BankPlugin"
        )
        self.object_bucket = "objects"
        s3_bank_plugin_cls = importutils.import_class(
            import_str=import_str)

        self.s3_bank_plugin = s3_bank_plugin_cls(CONF, None)

    def test_acquire_lease(self):
        self.s3_bank_plugin.acquire_lease()
        expire_time = math.floor(time.time()) + self.conf.lease_expire_window
        self.assertEqual(self.s3_bank_plugin.lease_expire_time, expire_time)

    def test_renew_lease(self):
        self.s3_bank_plugin.acquire_lease()
        expire_time = math.floor(time.time()) + self.conf.lease_expire_window
        self.assertEqual(self.s3_bank_plugin.lease_expire_time, expire_time)
        time.sleep(5)
        self.s3_bank_plugin.acquire_lease()
        expire_time = math.floor(time.time()) + self.conf.lease_expire_window
        self.assertEqual(self.s3_bank_plugin.lease_expire_time, expire_time)

    def test_check_lease_validity(self):
        self.s3_bank_plugin.acquire_lease()
        expire_time = math.floor(time.time()) + self.conf.lease_expire_window
        self.assertEqual(self.s3_bank_plugin.lease_expire_time, expire_time)
        is_valid = self.s3_bank_plugin.check_lease_validity()
        self.assertEqual(is_valid, True)

    def test_delete_object(self):
        self.s3_bank_plugin.update_object("key", "value")
        self.s3_bank_plugin.delete_object("key")
        object_list = self.s3_bank_plugin.list_objects()
        self.assertEqual('key' in object_list, False)

    def test_get_object(self):
        self.s3_bank_plugin.update_object("key", "value")
        value = self.s3_bank_plugin.get_object("key")
        self.assertEqual(value, "value")

    def test_list_objects(self):
        self.s3_bank_plugin.update_object("key-1", "value-1")
        self.s3_bank_plugin.update_object("key-2", "value-2")
        objects = self.s3_bank_plugin.list_objects(prefix=None)
        self.assertEqual(len(objects), 2)

    def test_update_object(self):
        self.s3_bank_plugin.update_object("key-1", "value-1")
        self.s3_bank_plugin.update_object("key-1", "value-2")
        contents = self.s3_bank_plugin.get_object('key-1')
        self.assertEqual(contents, "value-2")

    def test_create_get_dict_object(self):
        self.s3_bank_plugin.update_object("dict_object", {"key": "value"})
        value = self.s3_bank_plugin.get_object("dict_object")
        self.assertEqual(value, {"key": "value"})
